/* Copyright (C) 2000-2006 Lavtech.com corp. All rights reserved.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA 
*/

#include "udm_config.h"

#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_IO_H
#include <io.h>
#endif
#include <sys/stat.h>
#include <stdio.h>
#include <errno.h>
#include <math.h>

#include "udm_common.h"
#include "udm_utils.h"
#include "udm_unicode.h"
#include "udm_unidata.h"
#include "udm_uniconv.h"
#include "udm_searchtool.h"
#include "udm_boolean.h"
#include "udm_xmalloc.h"
#include "udm_spell.h"
#include "udm_stopwords.h"
#include "udm_word.h"
#include "udm_vars.h"
#include "udm_db.h"
#include "udm_db_int.h"
#include "udm_url.h"
#include "udm_hash.h"
#include "udm_parsehtml.h"
#include "udm_store.h"
#include "udm_doc.h"
#include "udm_conf.h"
#include "udm_result.h"
#include "udm_log.h"
#include "udm_sgml.h"
#include "udm_mutex.h"
#include "udm_chinese.h"
#include "udm_synonym.h"
#include "udm_sqldbms.h"

#ifdef CHASEN
#include <chasen.h>
#endif

#ifdef MECAB
#include <mecab.h>
#endif

/*
#define DEBUG_CACHE
*/

/********** QSORT functions *******************************/

static int cmpurlid (UDM_URL_CRD *s1, UDM_URL_CRD *s2)
{
  if (s1->url_id > s2->url_id) return(1);
  if (s1->url_id < s2->url_id) return(-1);
  if (s1->secno > s2->secno) return 1;
  if (s1->secno < s2->secno) return -1;
  return (int) s1->pos - (int) s2->pos;
}


static int
cmppattern(UDM_URLDATALIST *Data, UDM_URLDATA *D,
           long j, const char *pattern)
{
  int rc;

  for(; *pattern != '\0'; pattern++)
  {
    switch(*pattern)
    {
      case 'R':
      case 'r':
        if (D->coord > Data->Item[j].coord) return (*pattern == 'R') ? 1 : -1;
        if (D->coord < Data->Item[j].coord) return (*pattern == 'R') ? -1 : 1;
        break;
      case 'P':
      case 'p':
        if (D->pop_rank > Data->Item[j].pop_rank) return (*pattern == 'P') ? 1 : -1;
        if (D->pop_rank < Data->Item[j].pop_rank) return (*pattern == 'P') ? -1 : 1;
        break;
      case 'D':
      case 'd':
        if (D->last_mod_time > Data->Item[j].last_mod_time) return (*pattern == 'D') ? 1 : -1;
        if (D->last_mod_time < Data->Item[j].last_mod_time) return (*pattern == 'D') ? -1 : 1;
        break;
      case 'U':
      case 'u':
        rc= strcmp(UDM_NULL2EMPTY(D->url), UDM_NULL2EMPTY(Data->Item[j].url));
        if (rc) return(*pattern == 'U' ? -rc : rc);
        break;
      case 'S':
      case 's':
        rc= strcmp(UDM_NULL2EMPTY(D->section),
                   UDM_NULL2EMPTY(Data->Item[j].section));
        if (rc) return(*pattern == 'S' ? -rc : rc);
        break;
    }
  }
  return 0;
}


static int
cmppatternRP(UDM_URLDATA *D1, UDM_URLDATA *D2)
{
  if (D1->coord > D2->coord) return -1;
  if (D1->coord < D2->coord) return  1;
  
  if (D1->pop_rank > D2->pop_rank) return -1;
  if (D1->pop_rank < D2->pop_rank) return 1;
  return 0;
}


static int
cmppatternR(UDM_URLDATA *D1, UDM_URLDATA *D2)
{
  if (D1->coord > D2->coord) return -1;
  if (D1->coord < D2->coord) return  1;
  return 0;
}



/****************************************************/

void UdmSortSearchWordsByURL(UDM_URL_CRD *wrd,size_t num)
{
  if(wrd)UdmSort((void*)wrd,num,sizeof(*wrd),(udm_qsort_cmp)cmpurlid);
  return;
}


static int
cmpsiteid2(UDM_URLDATA *p1, UDM_URLDATA *p2)
{
  if (p1->site_id  > p2->site_id)  return -1;
  if (p1->site_id  < p2->site_id)  return 1;
#if 0
  if (p1->coord    > p2->coord)    return -1;
  if (p1->coord    < p2->coord)    return 1;
  if (p1->pop_rank > p2->pop_rank) return -1;
  if (p1->pop_rank < p2->pop_rank) return 1;
#endif
  return 0;
}


void
UdmSortURLDataBySite(UDM_URLDATALIST *L)
{
  qsort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmpsiteid2);
#if 0
  {
    size_t i;
    printf("L->num: %d num: %d\n", L->ncoords, num);
    for (i= 0; i < num; i++)
      printf("%d %d %d %d %.6f\n", i,
      L->Data[i].url_id, L->Data[i].site_id, L->Data[i].coord, L->Data[i].pop_rank);
  }
#endif
}

static void
UdmSortSearchWordsByPatternRP(UDM_URLDATALIST *L)
{
  qsort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmppatternRP);
}


static void
UdmSortSearchWordsByPatternR(UDM_URLDATALIST *L)
{
  qsort((void*) L->Item, L->nitems, sizeof(*L->Item), (udm_qsort_cmp) cmppatternR);
}


static size_t UdmH[] = {1, 5, 19, 41, 109, 209, 505, 929, 2161,
                        3905, 8929, 16001, 36289, 64769};
void UdmSortSearchWordsByPattern(UDM_URLDATALIST *L, const char *pattern)
{
  register ssize_t h, i, j;
  int s = 13;
  size_t num= L->nitems;
  UDM_URLDATA Dat;

  if (!strcmp(pattern, "RP"))
  {
    UdmSortSearchWordsByPatternRP(L);
    goto ret;
  }
  if (!strcmp(pattern, "R"))
  {
    UdmSortSearchWordsByPatternR(L);
    goto ret;
  }

  while( (s > 0) && ((num / 3) < UdmH[s])) s--;
  while(s >= 0)
  {
    h= UdmH[s];
    for (j= h; j < (ssize_t) num; j++)
    {
      Dat= L->Item[j];
      i= j - h;
D4:
      if (cmppattern(L, &Dat, i, pattern) <= 0) goto D6;
      L->Item[i + h]= L->Item[i];
      i-= h;
      if (i >= 0) goto D4;
D6:
      L->Item[i + h]= Dat;
    }
    s--;
  }

ret:

#if 0
  {
    size_t i;
    printf("L->num: %d num:%d pattern:%s\n", L->ncoords, num, pattern);
    for (i= 0; i < num; i++)
    {
      printf("[%d]%d %d %d %d %.6f\n", i,
             L->Coords[i].url_id, L->Coords[i].coord,
             L->Data[i].url_id, L->Data[i].coord, (float) L->Data[i].pop_rank);
    }
  }
#endif
  return;
}


/*#define DEBUG_TOP_SORT*/

/* Find topcount best results */
#if 0
void UdmWrdTopSort(UDM_URL_CRD *wrd, size_t nwrd,size_t topcount)
{
  size_t j;
  UDM_URL_CRD w;
  
#ifdef DEBUG_TOP_SORT
  fprintf(stderr,"top+1=%d nwrd=%d\n",topcount+1,nwrd);
#endif
  
  /*
  UdmSortSearchWordsByWeight(wrd,topcount+1);
  */
  for(j=topcount;j<nwrd;j++)
  {
    register int res;
    if (wrd[j].coord > wrd[topcount].coord) res = 1;
    else if (wrd[j].coord < wrd[topcount].coord) res = -1;
    else
/*    if(!(res=(wrd[j].coord-wrd[topcount].coord)))*/
      res = (wrd[topcount].url_id - wrd[j].url_id);
    
#ifdef DEBUG_TOP_SORT
fprintf(stderr,"(%d,%d) %d (%d,%d) %d\n",
    wrd[topcount].coord,wrd[topcount].url_id,topcount,
    wrd[j].coord,wrd[j].url_id,j);
#endif

    if(res>0)
    {
      size_t l,c,r;
      
      l=0;r=topcount;
      while(l<r)
      {
        c=(l+r)/2;
        if (wrd[c].coord > wrd[j].coord) res = 1;
        else if (wrd[c].coord < wrd[j].coord) res = -1;
        else
/*        if(!(res=(wrd[c].coord-wrd[j].coord)))*/
          res = (wrd[j].url_id - wrd[c].url_id);
        
        if(res>0)
        {
          l=c+1;
        }
        else
        {
          r=c;
        }
      }
      w=wrd[topcount];
      memmove(&wrd[r+1],&wrd[r],(topcount-r)*sizeof(*wrd));
      wrd[r]=wrd[j];
      wrd[j]=w;
    }
  }
}
#endif

#define UDM_MAX_FORMS 256
#define UDM_MAX_NORMS 64

/*
  All the following combinations should
  work and get as many uword forms as possible:

  1. uword doesn't exist in ispell, its synonym doesn't exist in ispell.
     This last combination should also work if no ispell dictionaries loaded.
     Just copy all synonyms into result.
  2. DONE: both norm(uword) and its synonym exist in ispell
  3. norm(uword) exists in ispell, its synonym doesn't exist in ispell.
  4. uword doesn't exist in ispell, its synonym exists in ispell.
*/

static UDM_WIDEWORDLIST *UdmAllForms1(UDM_AGENT *Indexer,
                                      UDM_WIDEWORDLIST *result,
                                      const UDM_WIDEWORD *uword)
{
  UDM_SPELLLISTLIST *SLL= &Indexer->Conf->Spells;
  UDM_AFFIXLISTLIST *ALL= &Indexer->Conf->Affixes;
  UDM_SYNONYMLIST   *SYN= &Indexer->Conf->Synonyms;
  char *Res[UDM_MAX_FORMS];
  char **ResCur= Res;
  char **ResEnd= Res + UDM_MAX_FORMS;
  char **R;
  UDM_AFFIXLIST *Al;  
  UDM_WIDEWORD w;
  UDM_CHARSET *lcs= Indexer->Conf->lcs;
  int sy= UdmVarListFindInt(&Indexer->Conf->Vars, "sy", 1);
  int sp= UdmVarListFindInt(&Indexer->Conf->Vars, "sp", 1);

  if (!sp)
    return NULL;

  for (Al= ALL->Item; Al < &ALL->Item[ALL->nitems]; Al++)
  {
    UDM_SPELLLIST *Sl;
    for (Sl= SLL->Item; Sl < &SLL->Item[SLL->nitems]; Sl++)
    {
      if (!strcasecmp(Al->lang, Sl->lang) && !strcasecmp(Al->cset, Sl->cset))
      {
        UDM_SPELL Norm[UDM_MAX_NORMS];
        UDM_SPELL *NormEnd= Norm + UDM_MAX_NORMS;
        UDM_SPELL *NormCur= Norm;
        UDM_SPELL *N;
        char tmp[256];
        char *word= uword->word;
        UDM_CONV scs_lcs, lcs_scs;

        UdmConvInit(&scs_lcs, Sl->cs, lcs, UDM_RECODE_HTML);
        UdmConvInit(&lcs_scs, lcs, Sl->cs, UDM_RECODE_HTML);
        
        if (lcs != Sl->cs)
        {
          size_t len= strlen(word);
          UdmConv(&lcs_scs, tmp, sizeof(tmp), word, len + 1);
          word= tmp;
        }
        
        NormCur+= UdmSpellNormalize(Sl, Al, word, NormCur, NormEnd-NormCur);
        
        if (sy && SYN->nsynonyms)
        {
          UDM_WIDEWORD ww;
          UDM_WIDEWORDLIST *syn;
          bzero((void*)&ww, sizeof(ww));
          ww.order= uword->order;
          ww.word= tmp;
          /* 
            Find synonyms for each normal form
            and add the found synonyms into normalized
            list for futher denormalization.
          */
          for (N= Norm; N < NormCur; N++)
          {
            ww.len= UdmConv(&scs_lcs,tmp,sizeof(tmp),N->word,strlen(N->word)+1) - 1;
            if ((syn= UdmSynonymListFind(SYN, &ww)))
            {
              UDM_WIDEWORD *W;
              for (W= syn->Word; W < syn->Word + syn->nwords; W++)
              {
                size_t nbytes= W->len + 1;
                UdmConv(&lcs_scs, tmp, sizeof(tmp), (char*) W->word, nbytes);
                if (NormCur < NormEnd)
                {
                  NormCur+= UdmSpellNormalize(Sl, Al, tmp, NormCur, NormEnd-NormCur);
                }
              }
              UdmWideWordListFree(syn);
              UdmFree(syn);
            }
          }
        }
        
        for (N= Norm ; N < NormCur; N++)
        {
          if (ResCur < ResEnd)
          {
            size_t cres= 1;
            *ResCur= UdmStrdup(N->word);
            cres+= UdmSpellDenormalize(Sl, Al, N, ResCur+1, ResEnd-ResCur-1);
            if (lcs != Sl->cs)
            {
              size_t i;
              for (i=0; i < cres; i++)
              {
                UdmConv(&scs_lcs, tmp, sizeof(tmp),
                        ResCur[i], strlen(ResCur[i])+1);
                UdmFree(ResCur[i]);
                ResCur[i]= UdmStrdup(tmp);
              }
            }
            ResCur+= cres;
          }
        }
      }
    }
  }
  
  bzero((void*)&w, sizeof(w));
  for (R=Res; R < ResCur; R++)
  {
    w.order= uword->order;
    w.phrpos= uword->phrpos;
    w.phrlen= uword->phrlen;
    w.count= 0;
    w.origin= UDM_WORD_ORIGIN_SPELL;
    w.word= *R;
    w.len= strlen(w.word);
    UdmWideWordListAdd(result, &w);
    UdmFree(*R);
  }
  return result;
}


static const char *translit_cyr_lat[]=
{
  "a",  "b",  "v",  "g",  "d",  "e",  "zh", "z",
  "i",  "j",  "k",  "l",  "m",  "n",  "o",  "p",
  "r",  "s",  "t",  "u",  "f",  "h",  "c",  "ch",
  "sh", "sch","`",  "y",  "'",  "`e", "yu", "ya",
  "",   "yo"
};


static const char *translit_lat_cyr[]=
{
  "&#x430;", "&#x431;", "&#x446;", "&#x434;",
  "&#x435;", "&#x444;", "&#x433;", "&#x445;",
  "&#x438;", "&#x439;", "&#x43a;", "&#x43b;",
  "&#x43c;", "&#x43d;", "&#x43e;", "&#x43f;",
  "&#x433;", "&#x440;", "&#x441;", "&#x442;",
  "&#x443;", "&#x432;", "&#x432;", "&#x43a;&#x441;",
  "&#x44b;", "&#x437;"
};


typedef struct udm_translit_complex_subst_st
{
  const char *from;
  const char *to;
} UDM_TRANSLIT_COMPLEX_SUBST;


static UDM_TRANSLIT_COMPLEX_SUBST translit_lat_cyr_complex[]=
{
#if NOT_YET
  {"`"  , "&#x44a;"},
  {"'"  , "&#x44c;"},
  {"`e" , "&#x44d;"},
#endif
  {"ch" , "&#x447;"},
  {"sch", "&#x449;"},
  {"ya" , "&#x44f;"},
  {"zh" , "&#x436;"},
  {"yo" , "&#x451;"},
  {"kh" , "&#x445;"},
  {"sh" , "&#x448;"},
#ifdef NOT_YET
  {"yu" , "&#x44e;"}, /* ambiguous: YERU + U, or YU   */
#endif
  {NULL, NULL}
};


typedef struct udm_translit_st
{
  size_t from;
  size_t to;
  const char * const *translit;
  UDM_TRANSLIT_COMPLEX_SUBST *complex;
} UDM_TRANSLIT_TABLE;


static UDM_TRANSLIT_TABLE tr_cyr_lat=
{
  0x430, 0x451, translit_cyr_lat, NULL
};


static UDM_TRANSLIT_TABLE tr_lat_cyr=
{
  0x61, 0x7A, translit_lat_cyr, translit_lat_cyr_complex
};


static int UdmAllFormsTranslit(UDM_AGENT *A, UDM_WIDEWORDLIST *result,
                               const UDM_WIDEWORD *uword,
                               const UDM_TRANSLIT_TABLE *tr)
{
  int *wrd, tword[128], tmp[128], *t, *te= tword + 128 - 2;
  int subst= 0;
  UDM_CHARSET *latin1= UdmGetCharSet("iso-8859-1");
  UDM_CONV l1_uni, lcs_uni;
  UdmConvInit(&l1_uni, latin1, &udm_charset_sys_int, UDM_RECODE_HTML);
  UdmConvInit(&lcs_uni, A->Conf->lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
  UdmConv(&lcs_uni, (char*) tmp, sizeof(tmp),
                    uword->word, strlen(uword->word) + 1);

  for (wrd= tmp, t= tword; wrd[0] && t < te; )
  {
    if (*wrd >= tr->from && *wrd <= tr->to)
    {
      const char *repl= NULL;
      size_t len;
      UDM_TRANSLIT_COMPLEX_SUBST *cmpl;
      for (cmpl= tr->complex; cmpl && cmpl->from; cmpl++)
      {
        size_t pos;
        const char *from= cmpl->from;
        for (pos=0; from[pos] && from[pos] == wrd[pos]; pos++);
        if (!from[pos])
        {
          repl= cmpl->to;
          wrd+= pos;
          break;
        }
      }
      if (!repl)
      {
        repl= tr->translit[*wrd - tr->from];
        wrd++;
      }
      
      len= strlen(repl);
      len= UdmConv(&l1_uni, (char*) t, (te - t) * sizeof(*t), repl, len);
      t+= len / sizeof(*t);
      *t= 0;
      subst++;
    }
    else
    {
      *t++= *wrd++;
    }
  }
  *t= 0;
  if (subst)
  {
    UDM_WIDEWORD w;
    char lcsword[128];
    UDM_CONV uni_lcs;
    size_t nbytes= (t - tword + 1) * sizeof(*t);
    bzero((void*)&w, sizeof(w));
    UdmConvInit(&uni_lcs, &udm_charset_sys_int, A->Conf->lcs, UDM_RECODE_HTML);
    UdmConv(&uni_lcs, lcsword, sizeof(lcsword), (const char *) tword, nbytes);
    w.order= uword->order;
    w.phrpos= uword->phrpos;
    w.phrlen= uword->phrlen;
    w.count= 0;
    w.origin= UDM_WORD_ORIGIN_SYNONYM;
    w.word= lcsword;
    w.len= strlen(w.word);
    UdmWideWordListAdd(result, &w);
  }
  return UDM_OK;
}


static int
UdmAllForms2(UDM_AGENT *Indexer,
             UDM_WIDEWORDLIST *result,
             UDM_WIDEWORD *uw)
{
  UdmAllForms1(Indexer, result, uw);
#ifdef HAVE_SQL
  {
    const char *sql= UdmVarListFindStr(&Indexer->Conf->Vars, "SQLWordForms", NULL);
    if (sql && Indexer->Conf->dbl.nitems)
    {
      char *word= uw->word;
      char buf[1024];
      size_t i, nrows;
      UDM_SQLRES SQLRes;
      UDM_WIDEWORD form;
      
      UdmBuildParamStr(buf, sizeof(buf), sql, &word, 1);
      if (UDM_OK != UdmSQLQuery(&Indexer->Conf->dbl.db[0], &SQLRes, buf))
        return UDM_ERROR;
      nrows= UdmSQLNumRows(&SQLRes);
      bzero((void*) &form, sizeof(UDM_WIDEWORD));
      form.order= uw->order;
      form.phrpos= uw->phrpos;
      form.phrlen= uw->phrlen;
      form.count= 0;
      form.origin= UDM_WORD_ORIGIN_SYNONYM;
      form.weight= 0;
      form.match= uw->match;
      for (i= 0; i < nrows; i++)
      {
        form.len= UdmSQLLen(&SQLRes, 0, i);
        form.word= (char*) UdmSQLValue(&SQLRes, 0, i);
        UdmWideWordListAdd(result, &form);
      }
      UdmSQLFree(&SQLRes);
    }
  }
#endif
  return UDM_OK;
}


static UDM_WIDEWORDLIST *UdmAllForms(UDM_AGENT *Indexer,
                                     UDM_WIDEWORDLIST *result,
                                     UDM_WIDEWORD *uword)
{
  UDM_WIDEWORDLIST *uwordsyn;
  
  /*
    Generate all possible word forms for uword.
  */
  UdmAllForms2(Indexer, result, uword);
  
  if (UdmVarListFindBool(&Indexer->Conf->Vars, "tl", 0))
  {
    UDM_TRANSLIT_TABLE *tbl[]= {&tr_cyr_lat, &tr_lat_cyr, NULL}, **cur;
    for (cur= tbl; *cur; cur++)
    {
      UDM_WIDEWORDLIST translit;
      UdmWideWordListInit(&translit);
      UdmAllFormsTranslit(Indexer, &translit, uword, *cur);
      if (translit.nwords)
      {
        UDM_WIDEWORD *ww= &translit.Word[0];
        UdmWideWordListAdd(result, ww);
        UdmAllForms2(Indexer, result, ww);
      }
      UdmWideWordListFree(&translit);
    }
  }

  if (UdmVarListFindBool(&Indexer->Conf->Vars, "Dehyphenate", 0))
  {
    UDM_WIDEWORD form;
    char tmpword[128], *s, *d;
    strcpy(tmpword, uword->word);
    for (s= uword->word, d= tmpword; ; s++)
    {
      *d++= *s;
      if (UdmAutoPhraseChar((unsigned char) *s))
        d--;
      
      if (!*d)
        break;
    }

    bzero((void*) &form, sizeof(UDM_WIDEWORD));
    form.order= uword->order;
    form.phrpos= uword->phrpos;
    form.phrlen= uword->phrlen;
    form.count= 0;
    form.origin= UDM_WORD_ORIGIN_SYNONYM;
    form.weight= 0;
    form.match= uword->match;
    form.word= tmpword;
    form.len= strlen(tmpword);
    UdmWideWordListAdd(result, &form);
  }
  
  if (!UdmVarListFindInt(&Indexer->Conf->Vars, "sy", 1))
    return result;
  /*
     Combination one: uword is possibly a normalized form.
     Find all uword synonyms and then process then through
     ispell to generate all word forms for the synonyms.
  */
  if ((uwordsyn= UdmSynonymListFind(&Indexer->Conf->Synonyms, uword)))
  {
    UDM_WIDEWORD *ww;
    UDM_CONV uni_lcs;
    UdmConvInit(&uni_lcs, &udm_charset_sys_int, Indexer->Conf->lcs, UDM_RECODE_HTML); 

    for (ww= uwordsyn->Word; ww < &uwordsyn->Word[uwordsyn->nwords]; ww++)
    {
      UdmWideWordListAdd(result, ww);
      UdmAllForms2(Indexer, result, ww);
    }
    UdmWideWordListFree(uwordsyn);
    UdmFree(uwordsyn);
  }

  return result;
}




static int
UdmStackItemAdd(UDM_STACKITEMLIST *List, UDM_STACK_ITEM *item)
{
  if (List->nitems >= List->mitems)
  {
    List->mitems+= UDM_MAXSTACK;
    if (!(List->items= (UDM_STACK_ITEM*)UdmRealloc(List->items,
                                                   List->mitems *
                                                   sizeof(UDM_STACK_ITEM))))
      return UDM_ERROR;
  }
  List->items[List->nitems]= item[0];
  List->nitems++;
  return UDM_OK;
}


typedef struct udm_stack_parser_state_st
{
  int secno;
  int secno_match_type;
  int next_word_match_type;
  int use_numeric_operators;
  int word_match;
  int nphrasecmd;
  int auto_phrase;
  int dehyphenate;
  int phrpos;
  const char *lang;
} UDM_STACK_PARSER_STATE;



static int
UdmStackItemAddCmd(UDM_RESULT *Res,
                   UDM_STACK_PARSER_STATE *state,
                   int *lex, size_t length)
{
  int rc= UDM_OK;
  size_t i;
  UDM_STACK_ITEM item;
  
  for (i = 0; i < length; i++)
  {
     switch(lex[i])
     {
       case '&':
       case '+':
         item.cmd= UDM_STACK_AND;
         state->next_word_match_type= state->word_match;
         break;
       case '|':
         item.cmd= UDM_STACK_OR;
         state->next_word_match_type= state->word_match;
         break;
       case '~':
         item.cmd= UDM_STACK_NOT;
         state->next_word_match_type= state->word_match;
         break;
       case '(':
         item.cmd= UDM_STACK_LEFT;
         state->next_word_match_type= state->word_match;
         break;
       case ')':
         item.cmd= UDM_STACK_RIGHT;
         state->next_word_match_type= state->word_match;
         break;
       case '"':
         item.cmd= UDM_STACK_PHRASE;
         state->next_word_match_type= state->word_match;
         state->nphrasecmd++;
         break;
       case '<':
         state->next_word_match_type= state->use_numeric_operators ?
                                      UDM_MATCH_NUMERIC_LT : state->word_match;
         break;
       case '>':
         state->next_word_match_type= state->use_numeric_operators ? 
                                      UDM_MATCH_NUMERIC_GT : state->word_match;
         break;
       default:
         if (state->auto_phrase && !UdmAutoPhraseChar(lex[i]))
         {
           int quot= '"';
           item.cmd= UDM_STACK_PHRASE;
           item.arg= state->secno_match_type;
           UdmStackItemAddCmd(Res, state, &quot, 1);
           state->auto_phrase= 0;
         }
         continue;
     }
     /* Ignore all operators if we are in a phrase,  except phrase end. */
    if (!(state->nphrasecmd % 2) || item.cmd == UDM_STACK_PHRASE)
    {
      item.arg= state->secno_match_type;
      rc= UdmStackItemAdd(&Res->ItemList, &item);
      Res->ItemList.ncmds++;
    }
  }
  return rc;
}


static int
UdmStackItemAddWord(UDM_AGENT *query, UDM_RESULT *Res,
                   UDM_STACK_PARSER_STATE *state,
                   size_t length, int *lt, char *wrd)
{
  UDM_WIDEWORD OWord;
  UDM_WIDEWORDLIST Forms;
  UDM_STACK_ITEM item;
  int origin, rc, lt_auto_phrase= UdmAutoPhraseChar(lt[0]);
  size_t phrlen= 0;

  if(Res->WWList.nuniq >= UDM_MAXWORDPERQUERY-1)
    return UDM_OK;

  if (state->nphrasecmd % 2) /* in phrase */
  {
    if (((state->auto_phrase && !lt_auto_phrase) || lt[0] == '"'))
    {
      /* End of auto- or non-auto-phrase*/
      phrlen= state->phrpos + 1;
    }
  }
  else /* not in phrase */
  {
    if (lt_auto_phrase)
    {
      /* Start of auto-phrase */
      int quot= '"';
      state->auto_phrase= 1;
      item.cmd= UDM_STACK_PHRASE;
      item.arg= 0;
      UdmStackItemAddCmd(Res, state, &quot, 1);
      phrlen= 0;
    }
    else
      phrlen= 1; /* Single word */
  }
  
  item.cmd= UDM_STACK_WORD;
  item.arg= Res->WWList.nuniq;
  rc= UdmStackItemAdd(&Res->ItemList, &item);


  /*
    Check stopword only when full word.
    Substring searches should not exclude them.
  */
  if(state->word_match == UDM_MATCH_FULL &&
     (UdmStopListFind(&query->Conf->StopWords, wrd, state->lang) ||
      query->Conf->WordParam.min_word_len > length ||
      query->Conf->WordParam.max_word_len < length))
  {
    origin= UDM_WORD_ORIGIN_STOP;
    Res->ItemList.items[Res->ItemList.nitems - 1].cmd= UDM_STACK_STOP;
  }
  else
  {
    origin= UDM_WORD_ORIGIN_QUERY;
  }

  bzero((void*) &OWord, sizeof(OWord));
  OWord.len= strlen(wrd);
  OWord.order= Res->WWList.nuniq;
  OWord.phrpos= state->phrpos;
  OWord.phrlen= phrlen;
  OWord.count= 0;
  OWord.word= wrd;
  OWord.origin = origin;
  OWord.match= state->next_word_match_type;
  OWord.secno= state->secno;
  OWord.phrwidth= state->dehyphenate ? 1 : 0;
  
  UdmWideWordListAdd(&Res->WWList, &OWord);
  if (state->nphrasecmd % 2)
    state->phrpos++;
      
  if (origin == UDM_WORD_ORIGIN_STOP)
    return UDM_OK;
      
  UdmWideWordListInit(&Forms);
  if(UdmAllForms(query,&Forms,&OWord))
  {
    UDM_WIDEWORD FWord;
    size_t frm;
    bzero((void*) &FWord, sizeof(FWord));
    for (frm= 0; frm < Forms.nwords ; frm++)
    {
      FWord.order= Res->WWList.nuniq;
      FWord.phrpos= state->phrpos;
      FWord.phrlen= phrlen;
      FWord.count= 0;
      FWord.word= Forms.Word[frm].word;
      FWord.len= Forms.Word[frm].len;
      FWord.origin = Forms.Word[frm].origin;
      FWord.match= state->next_word_match_type;
      FWord.secno= state->secno;
      UdmWideWordListAdd(&Res->WWList,&FWord);
/*    UdmLog(query, UDM_LOG_DEBUG, "Word form: [%d] %s", FWord.origin, wrd);*/
    }
  }
  UdmWideWordListFree(&Forms);
  Res->WWList.nuniq++;
  return rc;
}


int UdmPrepare(UDM_AGENT * query,UDM_RESULT *Res)
{
  UDM_CHARSET * browser_cs, * local_cs, *sys_int;
  int  ctype;
  int *ustr, *uend, *lt, *lex;
  size_t ulen;
  const char * txt = UdmVarListFindStr(&query->Conf->Vars,"q","");
  const char * qprev = UdmVarListFindStr(&query->Conf->Vars,"qprev","");
  char *ltxt;
  size_t wlen, llen, obytes;
  char *wrd;
  int *uwrd;
  UDM_CONV uni_lc, bc_uni, bc_lc;
  const char *lang;
  UDM_STACK_PARSER_STATE state;
  int *(*UniGetSepToken)(int *str, int *strend, int **last, int *ctype0);
  
  state.secno= 0;
  state.secno_match_type= 0;
  state.use_numeric_operators= UdmVarListFindBool(&query->Conf->Vars, "UseNumericOperators", 0);
  state.dehyphenate= UdmVarListFindBool(&query->Conf->Vars, "Dehyphenate", 0);
  state.nphrasecmd= 0;
  state.word_match= UdmMatchMode(UdmVarListFindStr(&query->Conf->Vars, "wm", "wrd"));
  state.next_word_match_type= state.word_match;
  state.lang= UdmVarListFindStr(&query->Conf->Vars, "g", NULL);
  state.auto_phrase= 0;
  state.phrpos= 0;
  UniGetSepToken= state.dehyphenate ? UdmUniGetSepToken2 :UdmUniGetSepToken;
  
  if ((wrd = (char*)UdmMalloc(query->Conf->WordParam.max_word_len * 12 + 1)) == NULL) return 0;
  if ((uwrd = (int*)UdmMalloc(sizeof(int) * (query->Conf->WordParam.max_word_len + 1))) == NULL) { UDM_FREE(wrd); return 0; }


  if (!(browser_cs = query->Conf->bcs))
    browser_cs=UdmGetCharSet("iso-8859-1");
  
  if(!(local_cs = query->Conf->lcs))
    local_cs=UdmGetCharSet("iso-8859-1");
  
  sys_int= &udm_charset_sys_int;
  
  UdmConvInit(&bc_uni,browser_cs,sys_int,UDM_RECODE_HTML);
  UdmConvInit(&uni_lc,sys_int,local_cs,UDM_RECODE_HTML);
  UdmConvInit(&bc_lc,browser_cs,local_cs,UDM_RECODE_HTML);
  
  ulen=strlen(txt);
  ustr=(int*)(UdmMalloc((sizeof(int))*(ulen+1)));
  obytes= UdmConv(&bc_uni,(char*)ustr,sizeof(ustr[0])*(ulen+1),txt,ulen+1);
  obytes-= sizeof(int);
  
  /* Create copy of query, converted into LocalCharset (for UdmTrack) */
  llen = ulen * 14 + 1;
  ltxt=(char*)UdmMalloc(llen);
  obytes= UdmConv(&uni_lc, ltxt, llen, (char*) ustr, obytes);
  ltxt[obytes]='\0';
  UdmVarListReplaceStr(&query->Conf->Vars,"q",ltxt);  /* "q-lc" was here */
  UDM_FREE(ltxt);
  
  llen = strlen(qprev);
  ltxt=(char*)UdmMalloc(llen*14+1);
  obytes= UdmConv(&bc_lc,ltxt,llen*14+1,qprev,llen);
  ltxt[obytes]='\0';
  UdmVarListReplaceStr(&query->Conf->Vars,"qprev",ltxt);
  UDM_FREE(ltxt);
  
  /* Parse query and build boolean search stack*/
  UdmUniStrToLower(ustr);
  switch(browser_cs->family)
  {
    case UDM_CHARSET_CHINESE_SIMPLIFIED:
    case UDM_CHARSET_CHINESE_TRADITIONAL: lang = "zh"; break;
    case UDM_CHARSET_JAPANESE: lang = "ja"; break;
    case UDM_CHARSET_THAI: lang = "th"; break;
    default: lang = "";
  }
  
  ustr= UdmUniSegment(query, ustr, lang);
  uend= ustr + UdmUniLen(ustr); /* TODO: get rid of it */

  lex= UniGetSepToken(ustr, uend, &lt , &ctype);
  for ( ;lex; lex= UniGetSepToken(NULL, uend, &lt, &ctype))
  {
    wlen=lt-lex;
    memcpy(uwrd, lex, (udm_min(wlen, query->Conf->WordParam.max_word_len)) * sizeof(int));
    uwrd[udm_min(wlen, query->Conf->WordParam.max_word_len)] = 0;
    UdmConv(&uni_lc, wrd, query->Conf->WordParam.max_word_len * 12,(char*)uwrd, sizeof(uwrd[0])*(wlen+1));
    UdmTrim(wrd, " \t\r\n");
    
    if (ctype == UDM_UNI_SEPAR)
    {
      UdmStackItemAddCmd(Res, &state, lex, wlen);
    } 
    else
    {
      UDM_VAR *Section;

      if (lt[0] == ':' || lt[0] == '=')
      {
        if ((Section= UdmVarListFind(&query->Conf->Sections, wrd)))
        {
          state.secno= Section->section;
          state.secno_match_type= lt[0];
          continue;
        }
        if (wlen > 5 && !strncmp(wrd, "secno", 5))
        {
          state.secno= atoi(wrd + 5);
          state.secno_match_type= lt[0];
          continue;
        }
      }
      
      UdmStackItemAddWord(query, Res, &state, wlen, lt, wrd);
    }
  }
  
  if (state.nphrasecmd & 1)
  {
    UDM_STACK_ITEM item;
    item.cmd= UDM_STACK_PHRASE;
    item.arg= 0;
    UdmStackItemAdd(&Res->ItemList, &item);
    Res->ItemList.ncmds++;
  }
  UDM_FREE(ustr);
  UDM_FREE(uwrd);
  UDM_FREE(wrd);
  Res->WWList.wm= state.word_match;
  return UDM_OK;
}


typedef struct udm_score_param_st
{
  unsigned int *R;
  unsigned int *D;
  size_t D_size;
  size_t ncosine;
  size_t nsections;
  float Rsum_factor;
  size_t dst_offs;
  size_t nwf_offs;
  size_t nwf_num;
  unsigned int dst_weight;
  UDM_WIDEWORD *Word;
  size_t nuniq;
  size_t nwords;
  int search_mode;
  size_t *count;
  size_t count_size;
  UDM_STACKITEMLIST ItemList;
  char wf[256];
  char wf2[256];
  char nwf[256];
  float *numword_factor;
  float *numdistinctword_factor;
  float max_coord_factor;
  unsigned int MinCoordFactor;
  int have_WordFormFactor;
  float WordFormFactor;
  float WordFormFactorReminder;
  int SaveSectionSize;
  float WordDensityFactor;
  float WordDensityFactorReminder;
} UDM_SCORE_PARAM;



/*
  R[i] and D[i] are in the range 0..64.
  ns is between 1..256
*/
static
inline uint4 UdmCalcCosineWeight(UDM_SCORE_PARAM *score_param,
                                 float nwords_factor)
{
  size_t i, num= score_param->ncosine;
  unsigned int *D= score_param->D;
  unsigned int *R= score_param->R;
  float res;

  size_t Dsum=  D[0] * D[0];
  size_t RDsum= R[0] * D[0];

  for (i=1; i < num; i++)
  {
    if (D[i])
    {
      Dsum+=  D[i] * D[i];
      RDsum+= R[i] * D[i];
    }
  }
  
  res= score_param->Rsum_factor * nwords_factor *
       (float) RDsum / sqrt(Dsum) + 0.5;

#if 0
  for (i=0; i < num; i++)
  {
    if (D[i])
      fprintf(stderr, "[%d]=%d %d\n", i, R[i], D[i]);
  }
  fprintf(stderr, "Rsum_factor: %g  Dsum: %d  RD: %d cos: %d\n",
          Rsum_factor, Dsum, RDsum, (uint4) res);
#endif
  
  return (uint4) res;
}


static unsigned int
CalcAverageWordDistance(UDM_WIDEWORD *Word,
                        UDM_URL_CRD *phr, size_t num,
                        unsigned int dst_weight);


static uint4
UdmCalcCosineWeightQuick(UDM_SCORE_PARAM *score_param,
                         UDM_URL_CRD *Crd, size_t ncoords,
                         char *wf2, float nwords_factor)
{
  float res;
  
  if (ncoords == 1)
  {
    size_t wrdsec= Crd->secno;
    size_t wrdnum= score_param->Word[Crd->num].order;
    size_t offs= score_param->nsections * wrdnum + wrdsec;
    res= score_param->Rsum_factor * nwords_factor *
         (float) (score_param->R[offs]) + 0.5;
  }
  else
  {
    size_t Dsum, RDsum, add;
    char *added= (char*) score_param->D;
  
    bzero(added, score_param->ncosine);

    if (score_param->nuniq > 1)
    {
      add= CalcAverageWordDistance(score_param->Word, Crd, ncoords, score_param->dst_weight);
      Dsum= add * add;
      RDsum= score_param->R[score_param->dst_offs] * add;
    }
    else
    {
      Dsum= 0;
      RDsum= 0;
    }
    
    if (score_param->nwords == 1)
    {
      size_t weight= score_param->Word[0].weight;
      for( ; ncoords; Crd++, ncoords--)
      {
        size_t wrdsec= Crd->secno;
        if (!added[wrdsec])
        {
          add= wf2[wrdsec] + weight;
          Dsum+= add*add;
          RDsum+= add * score_param->R[wrdsec];
          added[wrdsec]= 1;
        }
      }
    }
    else
    {
      for( ; ncoords; Crd++, ncoords--)
      {
        size_t wrdsec= Crd->secno;
        size_t wrdnum= score_param->Word[Crd->num].order;
        size_t offs= score_param->nsections * wrdnum + wrdsec;
        if (!added[offs])
        {
          add= wf2[wrdsec] + score_param->Word[wrdnum].weight;
          Dsum+= add*add;
          RDsum+= add * score_param->R[offs];
          added[offs]= 1;
        }
      }
    }
    res= score_param->Rsum_factor * nwords_factor *
          (float) RDsum / sqrt(Dsum) + 0.5;
  }
  return (uint4) res;
}



static int UdmOriginWeight(int origin)
{
  switch(origin)
  {
    case UDM_WORD_ORIGIN_QUERY: return 3;
    case UDM_WORD_ORIGIN_SPELL: return 1;
  }
  return 0;
}


static void UdmWideWordListSetOriginWeight(UDM_WIDEWORDLIST *WWList)
{
  size_t i;
  for (i=0; i < WWList->nwords; i++)
    WWList->Word[i].weight= UdmOriginWeight(WWList->Word[i].origin);
}


typedef struct udm_coord_st
{
  uint4 pos;
  uint4 ord;
} UDM_COORD;


static unsigned int
CalcAverageWordDistance(UDM_WIDEWORD *Word,
                        UDM_URL_CRD *phr, size_t num,
                        unsigned int dst_weight)
{
  size_t sum, np;
  UDM_COORD prev;
  UDM_COORD curr;
  UDM_COORD next;
  UDM_URL_CRD *last;

  if (num < 2)
    return 0;

  
  if (num == 2)
  {
    size_t num0= phr[0].num;
    size_t num1= phr[1].num;
    size_t ord0= Word[num0].order;
    size_t ord1= Word[num1].order;
    size_t pos0= UDM_WRDDOCPOS(phr[0].secno, phr[0].pos);
    size_t pos1= UDM_WRDDOCPOS(phr[1].secno, phr[1].pos);
    size_t res= (ord0 == ord1) ? 0 : pos1 > pos0 ? pos1 - pos0 : pos0 - pos1;
    return res > 0 ? (res - 1) * dst_weight / 255 : 0;
  }
  
  
  np= sum= 0;
  last= phr + num;
  
  prev.pos= UDM_WRDDOCPOS(phr->secno, phr->pos);
  prev.ord= Word[phr->num].order;
  phr++;
  
  curr.pos= UDM_WRDDOCPOS(phr->secno, phr->pos);
  curr.ord= Word[phr->num].order;
  phr++;
  
  for (; phr < last; phr++)
  {
    next.pos= UDM_WRDDOCPOS(phr->secno, phr->pos);
    next.ord= Word[phr->num].order;
    
    if (prev.ord == curr.ord)
    {
      if (curr.ord == next.ord)
      {
        /* w1 w1 w1 */
      }
      else
      {
        /* w1 w1 w2 */
      }
    }
    else
    {
      if (curr.ord == next.ord)
      {
        /* w1 w2 w2 */
      }
      else
      {
        uint4 diff1, diff2;
        diff1= curr.pos - prev.pos;
        diff2= next.pos - curr.pos;
        if (diff1 > 0) diff1--;
        if (diff2 > 0) diff2--;
        if (prev.ord == next.ord)
        {
          /* w1 w2 w1 */
          sum+= diff1 < diff2 ? diff1 : diff2;
          np++;
        }
        else
        {
          /* w1 w2 w3 */
          sum+= diff1 + diff2;
          np+= 2;
        }
      }
    }
    prev= curr;
    curr= next;
  }
  /* return (np) ? ((float) sum / np) : 0; */
  return (np) ? (sum * dst_weight / np / 255) : 0;
}

#if 0
static void
UdmPrintCoords(UDM_URL_CRD *Coords, size_t ncoords)
{
  size_t i;
  for (i=0 ; i < ncoords; i++)
    fprintf(stderr, "[%d] %d %d %d/%d %d\n", i,
                    Coords[i].url_id,
                    Coords[i].secno,
                    Coords[i].pos,
                    Coords[i].seclen,
                    Coords[i].num);
}
#endif

static inline void CheckPhrase(
       UDM_WIDEWORDLIST *WWL,
       UDM_STACK_ITEM *query, size_t nitems,
       UDM_URL_CRD *coords, size_t ncoords,
       size_t *count)
{
  size_t q;
  size_t start, end, arg, i;
  size_t rstart, rend;
  size_t j, d, delta;
  UDM_URL_CRD *prev;

  /* find opening phrase command */
  for (q= 0; q < nitems; q++)
  {
    int cmd_arg;
    if (query[q].cmd != UDM_STACK_PHRASE) continue;
    cmd_arg= query[q].arg;
    
    /* find closing phrase command */
    start= q + 1;
    for (end= start; end < nitems && query[end].cmd != UDM_STACK_PHRASE; end++);
    q= end;
    arg= 0;

    /* skip trailing stopwords for now */
    /* TODO: we have to check document length (for phrases like "word1 stopword1") */
    for (rstart= start; rstart < end && query[rstart].cmd == UDM_STACK_STOP; rstart++);
    for (rend= end; rend > rstart && query[rend].cmd == UDM_STACK_STOP; rend--);

    /* if phrase contains stopwords only, we assume this document is found */
    if (rstart == rend) arg= 1;
    else for (i= 0; i < ncoords; i++)
    {
      size_t ord_i= WWL->Word[coords[i].num].order;
      if (query[rstart].arg != ord_i) continue;
      prev = &coords[i];
      d= i + 1;
      delta= 1;

      for (j= rstart + 1; j < rend; j++)
      {
        if (query[j].cmd == UDM_STACK_STOP)
        {
          delta++;
          continue;
        }

        /* find coord for this word */
        while (d < ncoords &&
               (UDM_WRDDOCPOS(prev->secno, prev->pos) == UDM_WRDDOCPOS(coords[d].secno, coords[d].pos) ||
                (UDM_WRDDOCPOS(prev->secno, prev->pos) + delta == UDM_WRDDOCPOS(coords[d].secno,coords[d].pos) &&
                 WWL->Word[coords[d].num].order != query[j].arg))) d++;

        if (d == ncoords ||
            UDM_WRDDOCPOS(prev->secno,prev->pos) != UDM_WRDDOCPOS(coords[d].secno,coords[d].pos) - delta ||
            query[j].arg != WWL->Word[coords[d].num].order) break;

        delta= 1;
        prev= &coords[d];
      }
      if (j == rend)
      {
        if (cmd_arg == '=' && rend - rstart + 1 != coords[i].seclen)
          continue;
        
        arg= 1;
        break;
      }
    }

    for (i= rstart; i < rend; i++)
      if (query[i].cmd == UDM_STACK_WORD) count[query[i].arg]= arg;
  }
}

/*
  TODO: implement tf*idf as follows:
  w_ik = tf_ik * idf_k

  where

  T_k = term k in document D_i
  tf_ik = frequency of term T_k in document D_i
  idf_k = inversed document frequency of term T_k in the collection C:

  idf_k = log( N / n_k)

  N = total number of documents in the collection C (i.e. in the result)
  n_k = the number of documents in C that contain T_k

  QQ: how to combine several words and sections?
  QQ: "number of documents in C" should probably mean
      "number of documents in the current search result" in our case.
*/

/*
#define DEBUG_REL 1
*/

#if 0
static void UdmGroupByURLNewRel(UDM_RESULT *Res)
{
  urlid_t prev_id;
  size_t prev_num= 0;
  size_t N= 0;
  size_t M= Res->WWList.nuniq;
  size_t ndoc= 0; 
  size_t i;
  UDM_URL_CRD *Crd = Res->CoordList.Coords;
  float  TFi[UDM_MAXWORDPERQUERY+1];
  float   Ci[UDM_MAXWORDPERQUERY+1];
  float   Vi[UDM_MAXWORDPERQUERY+1];    
  float   Wi[UDM_MAXWORDPERQUERY+1];
  float   Di[UDM_MAXWORDPERQUERY+1];
    
  bzero(TFi, sizeof(TFi));
  bzero(Ci, sizeof(Ci));
  bzero(Vi, sizeof(Vi));
  bzero(Di, sizeof(Di));
  
  
  /* N - number of documents */
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      N++;
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }

#ifdef DEBUG_REL
    if (i < Res->CoordList.ncoords)
    {
      fprintf(stderr,"[%d]doc=%d sec=%d wrd=%d pos=%d\n",
              i,Crd[i].url_id,
              UDM_WRDSEC(Crd[i].coord),
              UDM_WRDNUM(Crd[i].coord),
              UDM_WRDPOS(Crd[i].coord));
    }
#endif
  }
    
    
  /* TFi */
  for (i=0; i < Res->CoordList.ncoords; i++)
  {
    TFi[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
  /* Add fictious word */
  TFi[M]= ((float) Res->CoordList.ncoords) / M;
  M++;
    
    
    
  /* Ci */
  for (i=0; i < M; i++)
  {
     Ci[i]= ((float) TFi[i]) / N;
#ifdef DEBUG_REL
     fprintf(stderr,"[%d] TDi=%d Ci=%f\n",i,TFi[i], Ci[i]);
#endif
  }
    
    
  /* Vi */
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    size_t k;
      
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      Di[M-1]= TFi[M-1]/N; /* fictious word */
      for (k=0; k < M; k++)
      {
        Vi[k]+= (Di[k] - Ci[k]) * (Di[k] - Ci[k]);
      }
      bzero(Di, sizeof(Di));
        
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }
    if (i < Res->CoordList.ncoords)
      Di[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
    
  for (i=0; i < M; i++)
  {
    Vi[i]/= (N-1);
    Wi[i]= Ci[i] ? TFi[i]*Vi[i]/Ci[i]/Ci[i] : 0;
#ifdef DEBUG_REL
    fprintf(stderr,"[%d] Vi=%f Wi=%f\n", i, Vi[i],Wi[i]);
#endif
  }
    
    
  bzero(Di,sizeof(Di));
  for (prev_id= Crd[0].url_id, i=0; i<= Res->CoordList.ncoords; i++)
  {
    size_t k;
    
    if ( (i==Res->CoordList.ncoords) || (prev_id != Crd[i].url_id))
    {
      float sumWiDi= 0;
      float sumDi2= 0;
      float sumWi2= 0;
      float res;
      size_t n= 1;
        
      Di[M-1]= TFi[M-1]/N;  /* fictious word */
       
      for (k=0; k < M; k++)
      {
        sumWiDi+= Wi[k] * Di[k] / (float)M;
        sumDi2+=  Di[k] * Di[k] / (float)M;
        sumWi2+=  Wi[k] * Wi[k];
        if (k + 1 < M)
          n*= Di[k];
      }
        
      res= sumWiDi / sqrt(sumDi2 * sumWi2);
      Crd[ndoc].url_id= prev_id;
      Crd[ndoc].coord= res * 100000;
        
#ifdef DEBUG_REL
      fprintf(stderr,"RES[%d]=%f %f %f\n",
        prev_id, res, sumWiDi, sqrt(sumDi2 * sumWi2));
#endif 
        
      ndoc++;
        
      bzero(Di, sizeof(Di));
      if (i < Res->CoordList.ncoords)
      {
        prev_id= Crd[i].url_id;
        prev_num= i;
      }
    }
      
    if (i <= Res->CoordList.ncoords)
      Di[UDM_WRDNUM(Crd[i].coord)]++;
  }
    
  Res->CoordList.ncoords= ndoc;
  return;
  
}
#endif


#define NUMWORD_FACTOR_SIZE 256
#define NUMWORD_FACTOR(nwf, n) (nwf[n >= NUMWORD_FACTOR_SIZE ? NUMWORD_FACTOR_SIZE - 1 : n])
#define MAXCOORD_FACTOR(factor, max_coord) ((1-((float)factor*(float)max_coord)))
#define MINCOORD_FACTOR(factor, coord) ((float) 0x1000 / (float) (0x1000 + factor * (coord - 1)))

static
void UdmNumWordFactorInit(float k, float *numword_factor, size_t ncoords)
{
  size_t i;
  if (k > 1) k= 1;
  if (k < 0) k= 0;
  for (i= 0; i < NUMWORD_FACTOR_SIZE; i++)
  {
    numword_factor[i]= (1-k) + ((float) i / ncoords)*k;
  }
}


static
void UdmNWFNormalize(unsigned int *D, char *nwf, size_t nwf_num)
{
  size_t i, n;
  /* Find how many sections we have */
  for (i=0, n= 0 ; i < nwf_num; i++)
  {
    if (D[i])
      n++;
  }
  
  for (i=0 ; i < nwf_num; i++)
  {
    /*
      Affect result weight when this section is the only one
    */
    D[i]= (D[i] && n <= 1) ? nwf[i] * 16 : 0;
  }
}


static void FillD(unsigned int *D,
                  UDM_WIDEWORDLIST *WWL, 
                  UDM_URL_CRD *Crd, size_t ncoords,
                  size_t nsections,
                  char *wf2, size_t nwf_num,
                  unsigned int *D_nwf_offs,
                  unsigned int *D_dst_offs,
                  unsigned int dst_weight)
{
  *D_dst_offs= CalcAverageWordDistance(WWL->Word, Crd, ncoords, dst_weight);
  for( ; ncoords; Crd++, ncoords--)
  {
    size_t wrdsec= Crd->secno;
    size_t wrdnum= WWL->Word[Crd->num].order;
    UDM_WIDEWORD *W= &WWL->Word[wrdnum];
    D[nsections * wrdnum + wrdsec]= wf2[wrdsec] + W->weight;
    if (nwf_num)
      D_nwf_offs[wrdsec]= 1;
  }
}


static void
UdmApplyDensityFactor(UDM_SCORE_PARAM *param,
                      UDM_URL_CRD *CrdFrom,
                      size_t phr_n,
                      float *nword_factor)
{
  UDM_URL_CRD *CrdEnd, *CrdCurr;
  uint4 seclen[256];
  uint4 seccnt[256];
  uint4 wfsum= 0;
  float density_factor= 0;
  size_t i;
  size_t nsections= param->nsections;
      
  bzero(&seclen, sizeof(uint4)*nsections);
  bzero(&seccnt, sizeof(uint4)*nsections);
  for (CrdCurr= CrdFrom, CrdEnd= CrdFrom + phr_n; CrdCurr < CrdEnd; CrdCurr++)
  {
    size_t secno= CrdCurr->secno;
    seclen[secno]= CrdCurr->seclen;
    seccnt[secno]++;
  }
  
  for (i=0; i < nsections; i++)
  {
    if (seclen[i])
    {
      float add= (float) param->wf2[i] * seccnt[i] / seclen[i];
      wfsum+= param->wf2[i];
      density_factor+= add;
    }
  }
  if (wfsum)
  {
    density_factor= (density_factor > wfsum) ? 1 : density_factor / wfsum;
    *nword_factor= param->WordDensityFactor * density_factor + 
                   param->WordDensityFactorReminder * nword_factor[0];
  }
}


static void
UdmGroupByURLLoop(UDM_RESULT *Res,
                  UDM_URLCRDLIST *CoordList,
                  UDM_URLSCORELIST *ScoreList,
                  UDM_SCORE_PARAM *score_param)
{
  UDM_URL_CRD *Crd= CoordList->Coords;
  UDM_URL_SCORE *CrdTo= ScoreList->Item;
  UDM_URL_CRD *CrdFrom, *CrdCurr;
  UDM_URL_CRD *CrdLast= CoordList->Coords + CoordList->ncoords;
  UDM_WIDEWORD *Res_WWList_Word= score_param->Word;
  unsigned int *D= score_param->D;
  size_t nsections= score_param->nsections;
  unsigned int dst_weight= score_param->dst_weight;
  unsigned int *D_dst_offs= D + score_param->dst_offs;
  unsigned int *D_nwf_offs= D + score_param->nwf_offs;
  size_t nwf_num= score_param->nwf_num;
  int search_mode= score_param->search_mode;
  size_t *count= score_param->count;
  size_t count_size= score_param->count_size;
  UDM_STACK_ITEM *temp_items= score_param->ItemList.items;
  size_t nitems= score_param->ItemList.nitems;
  char *wf2= score_param->wf2;
  char *nwf= score_param->nwf;
  size_t D_size= score_param->D_size;
  float *numword_factor= score_param->numword_factor;
  float *numdistinctword_factor= score_param->numdistinctword_factor;
  float max_coord_factor= score_param->max_coord_factor;
  unsigned int MinCoordFactor= score_param->MinCoordFactor;
  int have_WordFormFactor= score_param->have_WordFormFactor;
  float WordFormFactor= score_param->WordFormFactor;
  float WordFormFactorReminder= score_param->WordFormFactorReminder;

  for (CrdFrom= Crd ; CrdFrom < CrdLast ; CrdFrom= CrdCurr)
  {
    float nword_factor;
    size_t z, phr_n, form_phr_n;
    size_t max_coord= CrdFrom->pos;
    size_t min_coord= CrdFrom->pos;
    bzero((void*)count, count_size);
    
    for (CrdCurr= CrdFrom ;
         CrdCurr < CrdLast && CrdCurr->url_id == CrdFrom->url_id;
         CrdCurr++)
    {
      size_t CrdCurr_pos= CrdCurr->pos;
      count[Res_WWList_Word[CrdCurr->num].order]++;
      if (max_coord < CrdCurr_pos)
        max_coord= CrdCurr_pos;
      if (min_coord > CrdCurr_pos)
        min_coord= CrdCurr_pos;
    }
    
    phr_n= CrdCurr - CrdFrom;
    
    switch(search_mode)
    {
      case UDM_MODE_BOOL:
        CheckPhrase(&Res->WWList, temp_items, nitems, CrdFrom, phr_n, count);
        if(!UdmCalcBoolItems(temp_items, nitems, count))
          continue;
        break;
        
      case UDM_MODE_ALL:
        if (phr_n < Res->WWList.nuniq)
          continue;
        if (Res->WWList.nuniq > 1)
        {
          for (z = 0; z < Res->WWList.nuniq; z++) if (count[z] == 0) break;
          if (z < Res->WWList.nuniq && count[z] == 0) continue;
        }
    }
    nword_factor= NUMWORD_FACTOR(numword_factor, phr_n) *
                  MAXCOORD_FACTOR(max_coord_factor, max_coord) *
                  MINCOORD_FACTOR(MinCoordFactor, min_coord);

    if (have_WordFormFactor)
    {
      float k;
      unsigned int phr_n2= phr_n * 2;
      for (form_phr_n= 0, z= 0 ; z < phr_n; z++)
      {
        int spell_factor[5]= {0, 2, 1, 1, 1};
        form_phr_n+= spell_factor[Res_WWList_Word[CrdFrom[z].num].origin];
      }
      k= WordFormFactor + WordFormFactorReminder * (float) form_phr_n / phr_n2;
      nword_factor*= k;
    }

    if (CrdFrom->seclen)
    {
      UdmApplyDensityFactor(score_param, CrdFrom, phr_n, &nword_factor);
    }
    

    if (search_mode != UDM_MODE_ALL)
    {
      size_t nuniq= 0;
      for (z= 0; z < Res->WWList.nuniq; z++)
      {
        if (count[z])
          nuniq++;
      }
      if (nuniq < Res->WWList.nuniq)
        nword_factor= nword_factor * numdistinctword_factor[nuniq];
    }

    if (nwf_num)
    {
      bzero((void*)D, D_size);
      FillD(D, &Res->WWList, CrdFrom, phr_n,
            nsections, wf2, nwf_num, D_nwf_offs, D_dst_offs, dst_weight);
      UdmNWFNormalize(D_nwf_offs, nwf, nwf_num);
      CrdTo->score= UdmCalcCosineWeight(score_param, nword_factor);
    }
    else
    {
      CrdTo->score= UdmCalcCosineWeightQuick(score_param,
                                             CrdFrom, phr_n,
                                             wf2, nword_factor);
    }
    CrdTo->url_id= CrdFrom->url_id;
    CrdTo++;
  }
  ScoreList->nitems= CrdTo - ScoreList->Item;
}



static void
UdmGroupByURLLoop_OneWord_All_NoNWF(UDM_RESULT *Res,
                                    UDM_URLCRDLIST *CoordList,
                                    UDM_URLSCORELIST *ScoreList,
                                    UDM_SCORE_PARAM *score_param)
{
  UDM_URL_SCORE *CrdTo= ScoreList->Item;
  UDM_URL_CRD *CrdFrom, *CrdCurr;
  UDM_URL_CRD *CrdLast= CoordList->Coords + CoordList->ncoords;
  char *wf2= score_param->wf2;
  float *numword_factor= score_param->numword_factor;
  float max_coord_factor= score_param->max_coord_factor;
  float Rsum_factor= score_param->Rsum_factor;
  unsigned int MinCoordFactor= score_param->MinCoordFactor;
  char *added= (char*) score_param->D;
  size_t ncosine= score_param->ncosine;
  size_t weight= score_param->Word[0].weight;
  float numword_factor_1= NUMWORD_FACTOR(numword_factor, 1);
  float Rsum_factor_mul_nword_factor_1= Rsum_factor * numword_factor_1;
  unsigned int *R= score_param->R;

  bzero((void*) added, ncosine);
  
  for (CrdFrom= CoordList->Coords ; CrdFrom < CrdLast ; CrdFrom= CrdCurr)
  {
    urlid_t CrdFrom_url_id= CrdFrom->url_id;
    
    CrdCurr= CrdFrom + 1;
    if (CrdCurr >= CrdLast || CrdCurr->url_id != CrdFrom_url_id)
    {
      unsigned char secno= CrdFrom->secno;
      if (CrdFrom->seclen)
      {
        float nword_factor= numword_factor_1 *
                    MAXCOORD_FACTOR(max_coord_factor, CrdFrom->pos) *
                    MINCOORD_FACTOR(MinCoordFactor, CrdFrom->pos);

        UdmApplyDensityFactor(score_param, CrdFrom, 1, &nword_factor);
        CrdTo->score= Rsum_factor * nword_factor * (float) (R[secno]) + 0.5;
        CrdTo->url_id= CrdFrom_url_id;
        CrdTo++;
      }
      else
      {
        CrdTo->score= Rsum_factor_mul_nword_factor_1 *
                      MAXCOORD_FACTOR(max_coord_factor, CrdFrom->pos) *
                      MINCOORD_FACTOR(MinCoordFactor, CrdFrom->pos) *
                      (float) (R[secno]) + 0.5;
        CrdTo->url_id= CrdFrom_url_id;
        CrdTo++;
      }
    }
    else
    {
      unsigned short max_coord, min_coord;
      unsigned char max_secno, prev_secno;
      size_t ncoords, add, Dsum, RDsum;
      float nword_factor;

      max_coord= min_coord= CrdFrom->pos;
      max_secno= prev_secno= CrdFrom->secno;
      
      /* Add the first coord */
      add= wf2[max_secno] + weight;
      Dsum= add * add;
      RDsum= add * R[max_secno];
      added[max_secno]= 1;
      
      /* Add all other coords */
      do 
      {
        unsigned char secno= CrdCurr->secno;
        unsigned short CrdCurr_pos= CrdCurr->pos;
        if (prev_secno != secno && !added[secno])
        {
          add= wf2[secno] + weight;
          Dsum+= add * add;
          RDsum+= add * R[secno];
          added[secno]= 1;
          if (secno > max_secno)
            max_secno= secno;
        }
        if (max_coord < CrdCurr_pos)
          max_coord= CrdCurr_pos;
        else if (min_coord > CrdCurr_pos)
          min_coord= CrdCurr_pos;
        CrdCurr++;
      } while (CrdCurr < CrdLast && CrdCurr->url_id == CrdFrom_url_id);

      bzero((void*) added, max_secno + 1);
      
      ncoords= CrdCurr - CrdFrom;
      
      nword_factor= NUMWORD_FACTOR(numword_factor, ncoords) *
                    MAXCOORD_FACTOR(max_coord_factor, max_coord) *
                    MINCOORD_FACTOR(MinCoordFactor, min_coord);

      if (CrdFrom->seclen)
      {
        UdmApplyDensityFactor(score_param, CrdFrom, ncoords, &nword_factor);
      }
      
      CrdTo->score= Rsum_factor * nword_factor * (float) RDsum / sqrt(Dsum) + 0.5;
      CrdTo->url_id= CrdFrom_url_id;
      CrdTo++;
    }
  }
  ScoreList->nitems= CrdTo - ScoreList->Item;
}


static void
UdmScoreParamInit(UDM_SCORE_PARAM *prm,
                  UDM_AGENT *query,
                  UDM_DB *db,
                  UDM_RESULT *Res)
{
  size_t i;
  float DefWordDensityFactor;
  UDM_VARLIST *Vars= &query->Conf->Vars;
  bzero((void*) prm, sizeof(UDM_SCORE_PARAM));
  prm->nsections = UdmVarListFindInt(Vars, "NumSections", 256);
  prm->dst_offs= Res->WWList.nuniq * prm->nsections;
  prm->nwf_offs= Res->WWList.nuniq * prm->nsections + 1;
  prm->max_coord_factor= ((float)UdmVarListFindInt(Vars, "MaxCoordFactor", 255)) / 0xFFFFFF;
  prm->MinCoordFactor= UdmVarListFindInt(Vars, "MinCoordFactor", 0);
  prm->have_WordFormFactor= UdmVarListFindInt(Vars, "WordFormFactor", 255) != 255;
  prm->WordFormFactor= ((float)UdmVarListFindDouble(Vars, "WordFormFactor", 255)) / 255;
  prm->WordFormFactorReminder= 1 - prm->WordFormFactor;
  prm->SaveSectionSize= UdmVarListFindBool(Vars, "SaveSectionSize", 1);
  DefWordDensityFactor= prm->SaveSectionSize ? 25 : 0;
  prm->WordDensityFactor= ((float)UdmVarListFindDouble(Vars, "WordDensityFactor", DefWordDensityFactor)) / 256;
  prm->WordDensityFactorReminder= 1 - prm->WordDensityFactor;
  prm->dst_weight= (unsigned int) UdmVarListFindInt(Vars, "WordDistanceWeight", 255);
  UdmWeightFactorsInit2(prm->wf, Vars, &db->Vars);

  for (i= 0; i < 256; i++)
    prm->wf2[i]= prm->wf[i] << 2;

  if (UdmVarListFindStr(Vars, "nwf", NULL))
  {
    UdmWeightFactorsInit(prm->nwf, UdmVarListFindStr(Vars, "nwf", ""), prm->nsections);
    prm->nwf_num= prm->nsections;
  }  

  prm->ncosine= Res->WWList.nuniq * prm->nsections + 1 + prm->nwf_num;
  prm->D_size= prm->ncosine * sizeof(unsigned int);
}


static void
UdmNumDistinctWordFactorInit(unsigned int numdistinctwordfactor,
                             float *numdistinctword_factor, size_t nuniq)
{
  size_t i;
  for (i= 0; i < nuniq; i++)
  {
    unsigned int k= numdistinctwordfactor;
    float x= (float) i / nuniq;
    numdistinctword_factor[i]= ((x * x * x * x * k) + (256-k)) / 256;
  }
}


static void
UdmStackItemListFree(UDM_STACKITEMLIST *L)
{
  UDM_FREE(L->items);
}


static int
UdmStackItemListCopy(UDM_STACKITEMLIST *Dst,
                     UDM_STACKITEMLIST *Src, int search_mode)
{
  size_t i, j;
  int inphrase, add_cmd= UDM_STACK_AND;
  size_t mitems= (Src->nitems + 1) * 2;
  size_t nbytes= mitems * sizeof(UDM_STACK_ITEM);
  
  if (!(Dst->items= (UDM_STACK_ITEM*)UdmMalloc(nbytes)))
   return UDM_ERROR;  
  
  switch(search_mode)
  {
    case UDM_MODE_ANY:
      add_cmd = UDM_STACK_OR;
      break;
    case UDM_MODE_BOOL:
    case UDM_MODE_ALL:
      add_cmd = UDM_STACK_AND;
      break;
  }
  Dst->items[0]= Src->items[0];
  inphrase= (Src->items[0].cmd == UDM_STACK_PHRASE) ? 1 : 0;

  for (i= 1, j= 1; i < Src->nitems; i++)
  {
    /*
      If previous item is WORD or PHRASE or RIGHT or STOPWORD
      and next item is WORD or PHRASE or LEFT or STOPWORD
      and we are not in phrase
      we have to insert search mode dependent operator.
     */
    if ((Src->items[i - 1].cmd == UDM_STACK_WORD ||
         Src->items[i - 1].cmd == UDM_STACK_STOP ||
         Src->items[i - 1].cmd == UDM_STACK_PHRASE ||
         Src->items[i - 1].cmd == UDM_STACK_RIGHT) &&
        (Src->items[i].cmd == UDM_STACK_WORD ||
         Src->items[i].cmd == UDM_STACK_STOP ||
         Src->items[i].cmd == UDM_STACK_PHRASE ||
         Src->items[i].cmd == UDM_STACK_LEFT) &&
         !inphrase)
    {
      Dst->items[j].cmd = add_cmd;
      Dst->items[j].arg = 0;
      j++;
    }
    if (Src->items[i].cmd == UDM_STACK_PHRASE)
      inphrase= !inphrase;
    Dst->items[j++]= Src->items[i];
  }
  search_mode= UDM_MODE_BOOL;
  Dst->mitems= mitems;
  Dst->nitems= j;
  return UDM_OK;
}


static void
UdmGroupByURLInternal(UDM_AGENT *query,
                      UDM_RESULT *Res,
                      UDM_URLCRDLIST *CoordList,
                      UDM_URLSCORELIST *ScoreList,
                      UDM_SCORE_PARAM *score_param,
                      int search_mode)
{
  size_t  i, Rsum;
  UDM_WIDEWORD *Res_WWList_Word= Res->WWList.Word;
  float numword_factor[NUMWORD_FACTOR_SIZE];
  float numdistinctword_factor[NUMWORD_FACTOR_SIZE];

  float numwordfactor= ((float)UdmVarListFindDouble(&query->Conf->Vars,
                                                "NumWordFactor", 25.5)) / 255;
  unsigned int numdistinctwordfactor= (unsigned int) UdmVarListFindInt(&query->Conf->Vars,
                                                     "NumDistinctWordFactor", 0);

  if(!CoordList->ncoords) return;

  UdmNumWordFactorInit(numwordfactor, numword_factor, CoordList->ncoords);
  UdmNumDistinctWordFactorInit(numdistinctwordfactor, numdistinctword_factor, Res->WWList.nuniq);
  
  score_param->count_size= Res->WWList.nuniq * sizeof(size_t);
  score_param->count= (size_t*)UdmMalloc(score_param->count_size);
  score_param->R= (unsigned int*)UdmMalloc(score_param->D_size);
  score_param->D= (unsigned int*)UdmMalloc(score_param->D_size);
  if (!score_param->count || !score_param->R || !score_param->D)
    goto err;

  bzero((void*) score_param->R, score_param->D_size);
  
  for(Rsum=0, i= 0; i < Res->WWList.nwords; i++)
  {
    size_t secno;
    if (Res_WWList_Word[i].origin != UDM_WORD_ORIGIN_QUERY)
      continue;
    for (secno= 0; secno < score_param->nsections; secno++)
    {
      size_t offs= Res_WWList_Word[i].order * score_param->nsections + secno;
      score_param->R[offs] = score_param->wf2[secno] + Res_WWList_Word[i].weight;
      Rsum+= score_param->R[offs] * score_param->R[offs];
    }
  }
  
  Rsum+= score_param->R[score_param->dst_offs] * score_param->R[score_param->dst_offs];
  score_param->Rsum_factor= 100000 / sqrt(Rsum);
  
  if (Res->ItemList.ncmds > 0 || search_mode == UDM_MODE_BOOL)
  {
    if (UDM_OK != UdmStackItemListCopy(&score_param->ItemList,
                                       &Res->ItemList, search_mode))
      goto err;
    search_mode= UDM_MODE_BOOL;
  }

  score_param->Word= Res->WWList.Word;
  score_param->nuniq= Res->WWList.nuniq;
  score_param->nwords= Res->WWList.nwords;
  score_param->search_mode= search_mode;
  score_param->numword_factor= numword_factor;
  score_param->numdistinctword_factor= numdistinctword_factor;

  if (search_mode == UDM_MODE_ALL &&
      Res->WWList.nwords == 1 &&
      score_param->nwf_num == 0 &&
      !numdistinctwordfactor)
  {
    UdmGroupByURLLoop_OneWord_All_NoNWF(Res, CoordList, ScoreList, score_param);
  }
  else
  {
    UdmGroupByURLLoop(Res, CoordList, ScoreList, score_param);
  }

err:

  UdmStackItemListFree(&score_param->ItemList);
  UDM_FREE(score_param->D);
  UDM_FREE(score_param->R);
  UDM_FREE(score_param->count);
  return;
}


void UdmGroupByURL(UDM_AGENT *query,
                   UDM_DB *db,
                   UDM_RESULT *Res,
                   UDM_URLCRDLIST *CoordList,
                   UDM_URLSCORELIST *ScoreList)
{
  UDM_SCORE_PARAM *prm;
  int search_mode= UdmSearchMode(UdmVarListFindStr(&query->Conf->Vars, "m", "all"));
  size_t threshold= UdmVarListFindInt(&query->Conf->Vars, "StrictModeThreshold", 0);
  size_t ncoords= (search_mode == UDM_MODE_ALL) && threshold ?
                  CoordList->ncoords : 0;
  size_t nbytes;

  UdmWideWordListSetOriginWeight(&Res->WWList);
  
  if (!(prm= UdmMalloc(sizeof(UDM_SCORE_PARAM))))
    return;
  
  UdmScoreParamInit(prm, query, db, Res);

  nbytes= CoordList->ncoords * sizeof(UDM_URL_SCORE);
  ScoreList->Item= (UDM_URL_SCORE*) UdmMalloc(nbytes);
  UdmGroupByURLInternal(query, Res, CoordList, ScoreList, prm, search_mode);
  
  if (ncoords && (ScoreList->nitems < threshold))
  {
    size_t strict_mode_found= ScoreList->nitems;
    UdmLog(query, UDM_LOG_DEBUG,
          "Too few results: %d, Threshold: %d, group in ANY mode", 
          strict_mode_found, threshold);
    UdmGroupByURLInternal(query, Res, CoordList, ScoreList, prm, UDM_MODE_ANY);
    if (ScoreList->nitems > strict_mode_found)
      UdmVarListReplaceInt(&query->Conf->Vars, "StrictModeFound", strict_mode_found);
  }
  
  UdmFree(prm);
}


void UdmGroupURLDataBySite(UDM_URLDATALIST *List)
{
  UDM_URLDATA *src= List->Item + 1;
  UDM_URLDATA *dst= List->Item;
  UDM_URLDATA *end= List->Item + List->nitems;
  uint4 count;
  
  if (!List->nitems)
    return;

  for(count= List->Item[0].per_site; src < end; src++)
  {
    /* Group by site_id */
    if(dst->site_id == src->site_id)
    {
      count+= src->per_site;
      if (dst->coord > src->coord)
      {
        continue;
      }
      else if (dst->coord == src->coord)
      {
        if (dst->pop_rank > src->pop_rank)
        {
          continue;
        }
        else if (dst->pop_rank == src->pop_rank)
        {
          if (dst->url_id < src->url_id)
            continue;
        }
      }
      dst->url_id=        src->url_id;
      dst->coord=         src->coord;
      dst->last_mod_time= src->last_mod_time;
      dst->pop_rank=      src->pop_rank;
      dst->url=           src->url;
      dst->section=       src->section;
    }
    else
    {
      /* Next site */
      dst->per_site= count;
      *++dst= *src;
      count= src->per_site;
    }
  }
  dst->per_site= count;
  List->nitems= dst - List->Item + 1;
  return;
}


/******** Convert category string into 32 bit number *************/

void UdmDecodeHex8Str(const char *hex_str, uint4 *hi,
                      uint4 *lo, uint4 *fhi, uint4 *flo)
{
  char str[33],str_hi[17],str_lo[17], *s = str;

  strncpy(str, hex_str, 13);
  str[12] = '\0';
  strcat(str,"000000000000");
  while(*s == '0') *s++ = ' ';
  strncpy(str_hi,&str[0],6); str_hi[6]=0;
  strncpy(str_lo,&str[6],6); str_lo[6]=0;
  
  *hi = (uint4)strtoul(str_hi, (char **)NULL, 36);
  *lo = (uint4)strtoul(str_lo, (char **)NULL, 36);

  if ((fhi != NULL) && (flo != NULL))
  {
    strncpy(str, hex_str, 13);
    str[12] = '\0';
    strcat(str,"ZZZZZZZZZZZZ");
    strncpy(str_hi, &str[0], 6); str_hi[6] = 0;
    strncpy(str_lo, &str[6], 6); str_lo[6] = 0;
  
    *fhi = strtoul(str_hi, (char **)NULL, 36);
    *flo = strtoul(str_lo, (char **)NULL, 36);
  }
}

int __UDMCALL UdmAddSearchLimit(UDM_AGENT *Agent, int type,
                                const char *file_name, const char *val)
{
  uint4 hi, lo, f_hi, f_lo;
  
  if(Agent->nlimits == MAX_SEARCH_LIMIT - 1) return(1);
  
  Agent->limits[Agent->nlimits].type = type;
  strcpy(Agent->limits[Agent->nlimits].file_name, file_name);
  switch(type)
  {
    case 0: UdmDecodeHex8Str(val, &hi, &lo, &f_hi, &f_lo); break;
    case 1: f_hi = hi = 0; f_lo = lo = 0; break;
    case 2: hi=atoi(val); lo=0; f_hi = hi; f_lo = lo; break;
    case 3: hi=UdmStrHash32(val); lo = 0; f_hi = hi; f_lo = 0; break;
  }  
  Agent->limits[Agent->nlimits].hi = hi;
  Agent->limits[Agent->nlimits].lo = lo;
  Agent->limits[Agent->nlimits].f_hi = f_hi;
  Agent->limits[Agent->nlimits].f_lo = f_lo;
  
  Agent->nlimits++;
  
  UdmLog(Agent, UDM_LOG_DEBUG, "val: %s  %x %x   %x %x", val, hi, lo, f_hi, f_lo);
  
  return(0);
}

int UdmParseQueryString(UDM_AGENT * Agent,
                        UDM_VARLIST * vars,char * query_string)
{
  char * tok, *lt;
  size_t len;
  char *str = (char *)UdmMalloc((len = strlen(query_string)) + 7);
  char *qs = (char*)UdmStrdup(query_string);
  char qname[256];

  if ((str == NULL) || qs == NULL)
  {
    UDM_FREE(str);
    UDM_FREE(qs);
    return 1;
  }

  UdmSGMLUnescape(qs);
  
  tok = udm_strtok_r(qs, "&", &lt);
  while(tok)
  {
    char empty[]="";
    char * val;
    const char * lim;
    
    if((val=strchr(tok,'=')))
    {
      *val='\0';
      val++;
    }
    else
    {
      val=empty;
    }
    UdmUnescapeCGIQuery(str,val);
    UdmVarListAddQueryStr(vars,tok,str);
    udm_snprintf(qname, 256, "query.%s", tok);
    UdmVarListAddQueryStr(vars, qname, str);
    
    sprintf(str,"Limit-%s",tok);
    if((lim=UdmVarListFindStr(vars,str,NULL)))
    {
      int ltype=0;
      const char * type, * fname = NULL;
      char * llt;
      strncpy(str, lim, len);
      
      if((type = udm_strtok_r(str, ":", &llt)))
      {
        if(!strcasecmp(type, "category"))
        {
          ltype = UDM_LIMTYPE_NESTED; fname = UDM_LIMFNAME_CAT;
        }
        else if(!strcasecmp(type, "tag"))
        {
          ltype = UDM_LIMTYPE_LINEAR_CRC; fname = UDM_LIMFNAME_TAG;
        }
        else if(!strcasecmp(type, "time"))
        {
          ltype = UDM_LIMTYPE_TIME; fname = UDM_LIMFNAME_TIME;
        }
        else if(!strcasecmp(type, "hostname"))
        {
          ltype = UDM_LIMTYPE_LINEAR_CRC; fname = UDM_LIMFNAME_HOST;
        }
        else if(!strcasecmp(type, "language"))
        {
          ltype = UDM_LIMTYPE_LINEAR_CRC; fname = UDM_LIMFNAME_LANG;
        }
        else if(!strcasecmp(type, "content"))
        {
          ltype = UDM_LIMTYPE_LINEAR_CRC; fname = UDM_LIMFNAME_CTYPE;
        }
        else if(!strcasecmp(type, "siteid"))
        {
          ltype = UDM_LIMTYPE_LINEAR_INT; fname = UDM_LIMFNAME_SITE;
        }
        if((fname != NULL) && strlen(val))
        {
          UdmAddSearchLimit(Agent,ltype,fname,val);
        }
      }
    }
    tok = udm_strtok_r(NULL, "&", &lt);
  }
  UDM_FREE(str);
  UDM_FREE(qs);
  return 0;
}


#if 0
static UDM_WIDEWORD*
UdmWordInWWList(UDM_WIDEWORDLIST *List, int *tok, size_t toklen,
                int hlstop, size_t phrpos)
{
  size_t uw;
  
  for(uw=0; uw < List->nwords; uw++)
  {
    size_t slen;
    UDM_WIDEWORD *W= &List->Word[uw];
    if (W->phrpos != phrpos)
      continue;
    if (!hlstop && W->origin == UDM_WORD_ORIGIN_STOP)
      continue;
    slen= W->ulen;
    if (wordlen < slen)
      continue;
    if (wordlen == slen && !UdmUniStrNCaseCmp(tok, W->uword, slen))
      return W;
      
    if (wordlen > slen) switch (List->wm)
    {
      case UDM_MATCH_BEGIN:
        if (!UdmUniStrNCaseCmp(tok, W->uword, slen))
          return W;
        break;
      case UDM_MATCH_END:
        if (!UdmUniStrNCaseCmp(tok + wordlen - slen, W->uword, slen))
          return W;
        break;
      case UDM_MATCH_SUBSTR:
        {
          size_t l1, l2;
          for (l1 = 0; l1 < wordlen; l1++)
          {
            if (l1 + slen > wordlen) break;
            for (l2 = 0; l2 < slen; l2++)
            {
              if (UdmUniToLower(tok[l1 + l2]) != UdmUniToLower(W->uword[l2]))
                break;
            }
            if (l2 == slen)
            {
              return W;
              break;
            }
          }
        }
        break;
    }
  }
  return NULL;
}
#endif


static UDM_WIDEWORD*
UdmWordInWWList2(UDM_WIDEWORDLIST *List,
                 int *tok, size_t toklen,
                 UDM_CONV *uni_wcs, UDM_CONV *lc_uni, UDM_CONV *uni_bc,
                 int hlstop, size_t phrpos)
{
  size_t uw;
  char word[128]= "";
  size_t wordlen= 0;

  if (List->wm != UDM_MATCH_FULL)
  {
    wordlen= UdmConv(uni_wcs, word, sizeof(word) - 1,
                              (char*)tok, toklen*sizeof(int));
    word[wordlen]= '\0';
    UdmStrToLowerExt(uni_wcs->to, word, wordlen, UDM_RECODE_HTML);
  }
  
  for(uw=0; uw < List->nwords; uw++)
  {
    size_t W_len;
    UDM_WIDEWORD *W= &List->Word[uw];

    if (W->phrpos != phrpos)
      continue;
    if (!hlstop && W->origin == UDM_WORD_ORIGIN_STOP)
      continue;

    if (List->wm == UDM_MATCH_FULL)
    {
      if (!UdmStrCaseCmp2(uni_wcs, (const char*) tok, toklen*sizeof(int),
                                   W->word, W->len))
        return W;
    }

    W_len= W->len;

    if (wordlen < W_len)
      continue;
    if (wordlen == W_len && !memcmp(word, W->word, W_len))
      return W;

    if (wordlen > W_len) switch (List->wm)
    {
      case UDM_MATCH_BEGIN:
        if (!memcmp(word, W->word, W_len))
          return W;
        break;
      case UDM_MATCH_END:
        if (!memcmp(word + wordlen - W_len, W->word, W_len))
          return W;
        break;
      case UDM_MATCH_SUBSTR:
        if (strstr(word, W->word))
          return W;
        break;
      default:
        UDM_ASSERT(0); /* Inpossible */
    }
  }
  return NULL;
}


/*
  Remove hilight markers from a string and return
  its new length, in bytes.
*/
static size_t
UdmRemoveHl(UDM_CHARSET *cs, char *str, size_t from, size_t to)
{
  if (cs == &udm_charset_sys_int)
  {
    int *stri= (int*) str;
    int *s= stri + from / sizeof(int);
    int *e= stri + to / sizeof(int);
    int *dst= s;
    for (; s < e; s++)
    { 
      if (*s != 2 && *s != 3)
        *dst++= *s;
    }
    return (dst - stri) * sizeof(int);
  }
  else
  {
    char *s= str + from, *e= str + to, *dst= s;
    for (; s < e; s++)
    { 
      if (*s != 2 && *s != 3)
        *dst++= *s;
    }
    return dst - str;
  }
}


static size_t
UdmHlAppend(UDM_CONV *uni_bc, UDM_WIDEWORD *found,
            char *dst, size_t dstlen, size_t dstmaxlen,
            int *tok, size_t toklen)
{
  int i2= 2, i3= 3;

  if (found)
  {
    dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i2, sizeof(i2));
  }
  if (uni_bc->to == &udm_charset_sys_int)
  {
    memcpy(dst + dstlen, tok, sizeof(*tok) * toklen);
    dstlen+= sizeof(*tok) * toklen;
  }
  else
    dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) tok, sizeof(*tok) * toklen);

  if (found)
    dstlen+= UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i3, sizeof(i3));

  /*fprintf(stderr, "appended to '%.*s'\n", dstlen, dst);*/

  return dstlen;
}

/*
#define DEBUG_HL 0
*/

/* Returns a 0-terminated string */

size_t
UdmHlConvertExtWithConv(char *dst, size_t dstmaxlen,
                        UDM_WIDEWORDLIST *List,
                        const char *src, size_t srclen,
                        UDM_CONV *uni_wcs, UDM_CONV *lc_uni, UDM_CONV *uni_bc,
                        int hilight_stopwords)
{
  int		*tok, *lt, ctype, *uni, *uend;
  int           i0= 0;
  size_t        dstlen= 0, dstlen_phr= 0, ulen;
  size_t        unimaxlen, expected_phrpos= 0;

#ifdef DEBUG_HL
  fprintf(stderr, "wcs=%s fromcs=%s tocs=%s src='%s'\n",
          wcs->name, lc_uni->from->name, uni_bc->to->name, src);
#endif


  /* Convert to unicode */
  unimaxlen= (srclen + 10) * sizeof(int);
  uni= (int *)UdmMalloc(unimaxlen);
  ulen= UdmConv(lc_uni,(char*)uni, unimaxlen, src, srclen) / sizeof(int);
  uni[ulen]= '0';
  uend= uni + ulen;

  /* Parse unicode string */
  for (tok= UdmUniGetSepToken(uni, uend, &lt, &ctype) ; tok ;
       tok= UdmUniGetSepToken(NULL, uend, &lt, &ctype))
  {
    size_t toklen= lt - tok;

    if (ctype == UDM_UNI_SEPAR || !List)
    {
      dstlen= UdmHlAppend(uni_bc, NULL, dst, dstlen, dstmaxlen, tok, toklen);
    }
    else
    {
      UDM_WIDEWORD *found= UdmWordInWWList2(List, tok, toklen,
                                            uni_wcs, lc_uni, uni_bc,
                                            hilight_stopwords, expected_phrpos);
      dstlen= UdmHlAppend(uni_bc, found, dst, dstlen, dstmaxlen, tok, toklen);
      if (found)
      {
        if (found->phrpos + 1 == found->phrlen)
        {
          /* last in phrase found */
          expected_phrpos= 0;
          dstlen_phr= dstlen;
        }
        else
        {
          /* middle in phrase found */
          expected_phrpos++;
        }
      }
      else
      {
        /* No word found on expected phrase position, rollback */
        dstlen= UdmRemoveHl(uni_bc->to, dst, dstlen_phr, dstlen);
        dstlen_phr= dstlen;
        expected_phrpos= 0;
      }
    }
  }


#ifdef DEBUG_HL  
  fprintf(stderr, "end: expected_phrpos=%d dstlen=%d dstlen_phr=%d\n", expected_phrpos, dstlen, dstlen_phr);
#endif
  if (expected_phrpos > 0)
  {
    /* Roll back: incomplete last phrase */
    dstlen= UdmRemoveHl(uni_bc->to, dst, dstlen_phr, dstlen);
  }

  UdmConv(uni_bc, dst + dstlen, dstmaxlen, (char*) &i0, sizeof(i0));
  UdmFree(uni);
  return dstlen;
}

static char*
UdmHlConvertExtWithConvDup(UDM_WIDEWORDLIST *List,
                           const char *src, size_t srclen,
                           UDM_CONV *uni_wcs, UDM_CONV *lc_uni, UDM_CONV *uni_bc,
                           int hlstop)
{
  size_t dstlen;
  char *dst;
  if(!src || !srclen)return NULL;
  dstlen= srclen * 14 + 10;
  dst= (char*)UdmMalloc(dstlen);
  UdmHlConvertExtWithConv(dst, dstlen,
                          List, src, srclen,
                          uni_wcs, lc_uni, uni_bc, hlstop);
  return dst;
}


#ifdef NOT_YET
/*
  For character sets having "septok" and when lcs == bcs
  
  TODO: this function does not support HTML entities:
    &Auml;
    &#196;
    &#xC4;

  UdmWordInWWList2() should also be modified to unserstand entitites
  when searching words.
*/
static char *
UdmHlConvertExtQuick(UDM_WIDEWORDLIST *List, const char *src,
                     UDM_CHARSET *cs,
                     int hilight_stopwords)
{
  char        *dst;
  const char  *srcend, *tok, *lt;
  size_t      srclen, dstlen= 0, dstlen_phr= 0;
  size_t      dstmaxlen, expected_phrpos= 0;
  int ctype;

  if(!src || !(srclen = strlen(src)))
    return NULL;

  srcend= src + srclen;
  dstmaxlen= srclen * 14 + 10;
  dst= (char*)UdmMalloc(dstmaxlen + 1);

  for (tok= cs->septoken(cs, src, srcend, &lt, &ctype) ; tok ;
       tok= cs->septoken(cs, NULL, srcend, &lt, &ctype))
  {
    size_t toklen= lt - tok;

    if (ctype == UDM_UNI_SEPAR || !List || toklen > 127)
    {
      memcpy(dst + dstlen, tok, toklen);
      dstlen+= toklen;
    }
    else
    {
      UDM_WIDEWORD *found;
      {
        char tmp[128];
        memcpy(tmp, tok, toklen);
        tmp[toklen]= '\0';
        cs->lcase(cs, tmp, toklen);
        found= UdmWordInWWList2(List, tmp, toklen, hilight_stopwords, expected_phrpos);
      }

      if (dstlen + toklen + 2 >= dstmaxlen)
        break;
      
      if (found)
      {
        dst[dstlen++]= '\2';
        memcpy(dst + dstlen, tok, toklen);
        dstlen+= toklen;
        dst[dstlen++]= '\3';
      }
      else
      {
        memcpy(dst + dstlen, tok, toklen);
        dstlen+= toklen;
      }
      
      if (found)
      {
        if (found->phrpos + 1 == found->phrlen)
        {
          /* last in phrase found */
          expected_phrpos= 0;
          dstlen_phr= dstlen;
        }
        else
        {
          /* middle in phrase found */
          expected_phrpos++;
        }
      }
      else
      {
        /* No word found on expected phrase position, rollback */
        dstlen= UdmRemoveHl(cs, dst, dstlen_phr, dstlen);
        dstlen_phr= dstlen;
        expected_phrpos= 0;
      }
    }
  }

  if (expected_phrpos > 0)
  {
    /* Roll back: incomplete last phrase */
    dstlen= UdmRemoveHl(cs, dst, dstlen_phr, dstlen);
  }

  dst[dstlen]= '\0';
  return dst;
}
#endif


size_t UdmHlConvertExt(char *dst, size_t dstlen,
                       UDM_WIDEWORDLIST *List, UDM_CHARSET *wcs,
                       const char * src, size_t length,
                       UDM_CHARSET * lcs, UDM_CHARSET * bcs, int hlstop)
{
  UDM_CONV lc_uni, uni_bc, uni_wcs;
  UdmConvInit(&lc_uni, lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
  UdmConvInit(&uni_bc, &udm_charset_sys_int, bcs, UDM_RECODE_HTML);
  UdmConvInit(&uni_wcs, &udm_charset_sys_int, wcs, UDM_RECODE_HTML);
  return UdmHlConvertExtWithConv(dst, dstlen, List, src, length,
                                 &uni_wcs, &lc_uni, &uni_bc, hlstop);
}

static
char *UdmHlConvertExtDup(UDM_WIDEWORDLIST *List, UDM_CHARSET *wcs,
                         const char * src, size_t srclen,
                         UDM_CHARSET * lcs, UDM_CHARSET * bcs, int hlstop)
{
  UDM_CONV lc_uni, uni_bc, uni_wcs;
  UdmConvInit(&uni_wcs, &udm_charset_sys_int, wcs, UDM_RECODE_HTML);
  UdmConvInit(&lc_uni, lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
  UdmConvInit(&uni_bc, &udm_charset_sys_int, bcs, UDM_RECODE_HTML);
  return UdmHlConvertExtWithConvDup(List, src, srclen,
                                    &uni_wcs, &lc_uni, &uni_bc, hlstop);
}


/* For PHP module compatibility */
char * UdmHlConvert(UDM_WIDEWORDLIST *List,const char * src,
                    UDM_CHARSET * lcs, UDM_CHARSET * bcs)
{
  return UdmHlConvertExtDup(List, lcs, src, strlen(src), lcs, bcs, 1);
}


int UdmConvert(UDM_ENV *Conf, UDM_RESULT *Res,
               UDM_CHARSET *lcs, UDM_CHARSET *bcs)
{
  size_t i;
  UDM_CONV lc_bc, lc_uni, uni_bc;
  int hlstop= UdmVarListFindBool(&Conf->Vars, "ExcerptStopword", 1);

  /* Init converters */
  UdmConvInit(&lc_bc,lcs,bcs,UDM_RECODE_HTML);
  UdmConvInit(&lc_uni, lcs, &udm_charset_sys_int, UDM_RECODE_HTML);
  UdmConvInit(&uni_bc, &udm_charset_sys_int, bcs, UDM_RECODE_HTML);
  
  /* Convert word list */
  for(i=0;i<Res->WWList.nwords;i++)
  {
    UDM_WIDEWORD *W=&Res->WWList.Word[i];
    size_t len= strlen(W->word);
    char *newval= (char*)UdmMalloc(len * 12 + 1);

    len= UdmConv(&lc_bc,newval,len*12+1,W->word,len);
    newval[len]= '\0';
    UDM_FREE(W->word);
    W->word=newval;
    W->len= len;
  }
  
  /* Convert document sections */
  for(i=0;i<Res->num_rows;i++)
  {
    UDM_DOCUMENT  *D=&Res->Doc[i];
    size_t    sec;
    
    for(sec=0; sec < D->Sections.nvars; sec++)
    {
      UDM_VAR *Var= &D->Sections.Var[sec];
      
      /*
         A temporary fix to skip URL and CachedCopy:
         We will skip these sections for now.
         But this need a further fix in search.htm
         to distinguish two HTML formats:
         - HTML with &<>" escaped to &amp;&lt;&gt;&quot;
         - HTML with &<>" printed as is, no word hilight
         - HTML with &<>" printed as is, search word hilighted.
      */
      
      if (strcasecmp(Var->name,"URL") &&
          strcasecmp(Var->name,"CachedCopy") &&
          strcasecmp(Var->name,"Content-Type"))
      {
        char *newval;
        newval= UdmHlConvertExtWithConvDup(&Res->WWList,
                                           Var->val, Var->curlen,
                                           &uni_bc, &lc_uni, &uni_bc, hlstop);
        UDM_FREE(Var->val);
        Var->val= newval;
      }
    }
  }
  
  /* Convert Env->Vars */
  for (i = 0; i < Conf->Vars.nvars; i++)
  {
    UDM_VAR *Var= &Conf->Vars.Var[i];
    if (UdmVarType(Var) == UDM_VAR_STR &&
        strcasecmp(Var->name, "HlBeg") &&
        strcasecmp(Var->name, "HlEnd"))
    {
      size_t len= strlen(Var->val);
      char *newval= (char*)UdmMalloc(len * 12 + 1);
    
      UdmConv(&lc_bc, newval, len * 12 + 1, Var->val, len + 1);
      UDM_FREE(Var->val);
      Var->val= newval;
    }
  }
  
  return UDM_OK;
}


static char rm_hl_special[256]=
{
/*00*/  1,0,1,1,0,0,0,0,0,0,0,0,0,0,0,0,
/*10*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*20*/  0,0,0,0,0,0,1,0,0,0,0,0,0,0,0,0,  /*  !"#$%&'()*+,-./ */
/*30*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* 0123456789:;<=>? */
/*40*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* @ABCDEFGHIJKLMNO */
/*50*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* PQRSTUVWXYZ[\]^_ */
/*60*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* `abcdefghijklmno */
/*70*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  /* pqrstuvwxyz{|}~  */
/*80*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*90*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*A0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*B0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*C0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*D0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*E0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,
/*F0*/  0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};


char* UdmRemoveHiLightDup(const char *s)
{
  size_t len=strlen(s)+1;
  char   *d, *res = (char*)UdmMalloc(len);
  
  for(d= res; ; s++)
  {
    if ((((unsigned char) *s) <= '&') &&
        rm_hl_special[(unsigned char) *s])
    {
      switch(s[0])
      {
        case '\0':
          goto ex;
        case '\2':
        case '\3':
          break;
        case '&':
          if (s[1] == '#')
          {
            const char *e;
            int code= 0;
          
            for (e= s+2; (*e >= '0') && (*e <= '9'); code= code*10 + e[0]-'0', e++);
            if (*e == ';')
            {
              *d++= (code < 128) ? code : '?';
              s= e;
              break;
            }
          }
          /* pass through */
        
        default:
          *d++=*s;
      }
    }
    else
      *d++= *s;
  }
ex:
  *d='\0';
  return res;
}



int UdmCatToTextBuf(UDM_CATEGORY *C, char *textbuf, size_t len)
{
  char  *end;
  size_t  i;
  
  textbuf[0]='\0';
  
  end = textbuf;
  
  for(i = 0; i < C->ncategories; i++)
  {
    udm_snprintf(end, len - strlen(textbuf),
                 "<CAT\tid=\"%d\"\tpath=\"%s\"\tlink=\"%s\"\tname=\"%s\">\r\n",
                 C->Category[i].rec_id, C->Category[i].path,
                 C->Category[i].link, C->Category[i].name);
    end = end + strlen(end);
  }
  return UDM_OK;
}

int UdmCatFromTextBuf(UDM_CATEGORY *C, char *textbuf)
{
  const char  *htok, *last;
  UDM_HTMLTOK  tag;
  size_t    i, c;
  
  if (textbuf == NULL) return UDM_OK;
  UdmHTMLTOKInit(&tag);
  
  htok=UdmHTMLToken(textbuf,&last,&tag);
  
  if(!htok || tag.type != UDM_HTML_TAG)
    return UDM_OK;

  C->Category = (UDM_CATITEM*)UdmRealloc(C->Category, sizeof(UDM_CATITEM) * ((c = C->ncategories) + 1));
  bzero((void*)&C->Category[c], sizeof(UDM_CATITEM));
  
  for(i = 1; i < tag.ntoks; i++)
  {
    size_t  nlen = tag.toks[i].nlen;
    size_t  vlen = tag.toks[i].vlen;
    char  *name = UdmStrndup(tag.toks[i].name, nlen);
    char  *data = UdmStrndup(tag.toks[i].val, vlen);

    if (!strcmp(name, "id"))
    {
      C->Category[c].rec_id = atoi(data);
    }
    else if (!strcmp(name, "path"))
    {
      strncpy(C->Category[c].path, data, 128);
    }
    else if (!strcmp(name, "link"))
    {
      strncpy(C->Category[c].link, data, 128);
    }
    else if (!strcmp(name, "name"))
    {
      strncpy(C->Category[c].name, data, 128);
    }

    UDM_FREE(name);
    UDM_FREE(data);
  }

  C->ncategories++;
  return UDM_OK;
}

int *UdmUniSegment(UDM_AGENT *Indexer, int *ustr, const char *lang)
{
  size_t       dstlen = UdmUniLen(ustr);
  const char   *seg=  UdmVarListFindStr(&Indexer->Conf->Vars, "Segmenter", NULL);

#ifdef CHASEN
  if ((!seg  || !strcasecmp(seg, "Chasen")) &&
      (!lang || !strncasecmp(lang, "ja", 2)))
  {
    char        *eucstr, *eucstr_seg;
    UDM_CHARSET *eucjp_cs;
    UDM_CONV    uni_eucjp, eucjp_uni;
    size_t      reslen;
    
    eucjp_cs = UdmGetCharSet("euc-jp");
    if (!eucjp_cs) eucjp_cs = &udm_charset_sys_int;
    UdmConvInit(&uni_eucjp, &udm_charset_sys_int, eucjp_cs, UDM_RECODE_HTML);
    UdmConvInit(&eucjp_uni, eucjp_cs, &udm_charset_sys_int, UDM_RECODE_HTML);
    eucstr = (char*)UdmMalloc(12 * dstlen + 1);
    UdmConv(&uni_eucjp, eucstr, 12 * dstlen + 1, (char*)ustr, sizeof(*ustr)*(dstlen + 1));
    
    UDM_GETLOCK(Indexer, UDM_LOCK_SEGMENTER);
    eucstr_seg = chasen_sparse_tostr(eucstr);
    UDM_RELEASELOCK(Indexer, UDM_LOCK_SEGMENTER);
    
    reslen = strlen(eucstr_seg) + 1;
    ustr = (int*)UdmRealloc(ustr, reslen * sizeof(int));
    UdmConv(&eucjp_uni, (char*)ustr, reslen * sizeof(int), eucstr_seg, reslen);
    UDM_FREE(eucstr);
    dstlen = UdmUniLen(ustr);
    return ustr;
  }
#endif


#ifdef MECAB
  if ((!seg  || !strcasecmp(seg, "Mecab")) &&
      (!lang || !strncasecmp(lang, "ja", 2)))
  {
    UDM_CHARSET *sjis_cs;
    UDM_CONV    uni_sjis, sjis_uni;
    char        *sjisstr, *sjisstr_seg;
    size_t      reslen;

    sjis_cs = UdmGetCharSet("euc-jp");
    if (!sjis_cs) sjis_cs = &udm_charset_sys_int;
    UdmConvInit(&uni_sjis, &udm_charset_sys_int, sjis_cs, UDM_RECODE_HTML);
    UdmConvInit(&sjis_uni, sjis_cs, &udm_charset_sys_int, UDM_RECODE_HTML);

    sjisstr = (char*)UdmMalloc(12 * dstlen + 1);
    UdmConv(&uni_sjis, sjisstr, 12 * dstlen + 1, (char*)ustr, sizeof(*ustr) * (dstlen + 1));

    UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
#ifdef HAVE_PTHREADS
    mecab_lock(Indexer->Conf->mecab);
#endif
    sjisstr_seg = mecab_sparse_tostr(Indexer->Conf->mecab, sjisstr);
#ifdef HAVE_PTHREADS
    mecab_unlock(Indexer->Conf->mecab);
#endif
    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);

    reslen = strlen(sjisstr_seg) + 1;
    ustr = (int*)UdmRealloc(ustr, reslen * sizeof(int));
    UdmConv(&sjis_uni, (char*)ustr, reslen * sizeof(int), sjisstr_seg, reslen);
    UDM_FREE(sjisstr);
    dstlen = UdmUniLen(ustr);
    return ustr;
  }
#endif


#ifdef HAVE_CHARSET_gb2312
  if ((!seg  || !strcasecmp(seg, "Freq")) && Indexer->Conf->Chi.nwords &&
      (!lang || !lang[0] || 
       !strncasecmp(lang, "zh", 2) ||
       !strncasecmp(lang, "cn", 2)))
  {
    int *seg_ustr;
    UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
    seg_ustr = UdmSegmentByFreq(&Indexer->Conf->Chi, ustr);
    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
    if (seg_ustr != NULL)
    {
      UDM_FREE(ustr);
      ustr = seg_ustr;
    }
    dstlen = UdmUniLen(ustr);
    return ustr;
  }
#endif

  if ((!seg  || !strcasecmp(seg, "Freq")) && Indexer->Conf->Thai.nwords &&
      (!lang || !strncasecmp(lang, "th", 2)))
  {
    int *seg_ustr;
    UDM_GETLOCK(Indexer, UDM_LOCK_CONF);
    seg_ustr = UdmSegmentByFreq(&Indexer->Conf->Thai, ustr);
    UDM_RELEASELOCK(Indexer, UDM_LOCK_CONF);
    if (seg_ustr != NULL)
    {
      UDM_FREE(ustr);
      ustr = seg_ustr;
    }
    dstlen = UdmUniLen(ustr);
    return ustr;
  }

  return ustr;
}


static int
UdmWWListWordInfo(UDM_VARLIST *Vars, UDM_WIDEWORDLIST *WWList)
{
  size_t  len, i, j, wsize;
  char  *wordinfo= NULL, *end;
  size_t  corder= (size_t)-1, ccount= 0;
  int have_suggestions= 0;
  
  for(len = i = 0; i < WWList->nwords; i++) 
    len += WWList->Word[i].len + 64;
  
  wsize= (1+len)*sizeof(char);
  wordinfo= (char*) UdmMalloc(wsize);
  *wordinfo= '\0';
  
  UdmVarListAddInt(Vars, "nwords", WWList->nwords);

  for(i = 0; i < WWList->nwords; i++)
  {
    char name[32], count[32];
    if (WWList->Word[i].origin == UDM_WORD_ORIGIN_QUERY ||
        WWList->Word[i].origin == UDM_WORD_ORIGIN_SPELL ||
        WWList->Word[i].origin == UDM_WORD_ORIGIN_SYNONYM)
    {
      if(wordinfo[0])
        strcat(wordinfo,", ");
      sprintf(UDM_STREND(wordinfo)," %s : %d", WWList->Word[i].word, WWList->Word[i].count);
      sprintf(count, "%d", WWList->Word[i].count);
    }
    else if (WWList->Word[i].origin == UDM_WORD_ORIGIN_STOP)
    {
      if(wordinfo[0])
        strcat(wordinfo,", ");
      sprintf(UDM_STREND(wordinfo)," %s : stopword", WWList->Word[i].word);
      strcpy(count, "stopword");
    }
    sprintf(name, "word%d.word", i);
    UdmVarListAddStr(Vars, name, WWList->Word[i].word);
    sprintf(name, "word%d.count", i);
    UdmVarListAddStr(Vars, name, count);
    sprintf(name, "word%d.order", i);
    UdmVarListAddInt(Vars, name, WWList->Word[i].order);
  }
  
  UdmVarListReplaceStr(Vars, "WE", wordinfo);
  
  *wordinfo = '\0';
  for(i = 0; i < WWList->nwords; i++)
  {
    corder= WWList->Word[i].order;
    ccount= 0;
    for(j= 0; j < WWList->nwords; j++)
      if (WWList->Word[j].order == corder)
        ccount += WWList->Word[j].count;
    if (WWList->Word[i].origin == UDM_WORD_ORIGIN_STOP)
    {
      sprintf(UDM_STREND(wordinfo),"%s%s : stopword", (*wordinfo) ? ", " : "",  WWList->Word[i].word);
    }
    else if (WWList->Word[i].origin == UDM_WORD_ORIGIN_QUERY)
    {
      sprintf(UDM_STREND(wordinfo),"%s%s : %d / %d", 
        (*wordinfo) ? ", " : "", WWList->Word[i].word, WWList->Word[i].count, ccount);
    }
  }
  UdmVarListReplaceStr(Vars, "W", wordinfo);
  
  *wordinfo= '\0';
  end= wordinfo;
  for (i= 0; i < WWList->nwords; i++)
  {
    UDM_WIDEWORD *Wi= &WWList->Word[i];
    UDM_WIDEWORD *Wb= NULL;
   
    if (Wi->origin == UDM_WORD_ORIGIN_QUERY)
    {
      if (Wi->count > 0)
      {
        Wb= Wi;
      }
      else
      {
        ccount= 0;
        for (j= 0; j < WWList->nwords; j++)
        {
          UDM_WIDEWORD *Wj= &WWList->Word[j];
          if (Wj->origin == UDM_WORD_ORIGIN_SUGGEST &&
              Wj->order == Wi->order && Wj->count > ccount)
          {
            ccount= WWList->Word[j].count;
            Wb= Wj;
            have_suggestions= 1;
          }
        }
      }
    }
    else if (Wi->origin == UDM_WORD_ORIGIN_STOP)
    {
      Wb= Wi;
    }
    
    if (Wb)
    {
      sprintf(end, "%s%s", wordinfo[0] ? " " : "", Wb->word);
      end= end + strlen(end);
    }
  }
  
  if (have_suggestions)
    UdmVarListReplaceStr(Vars, "WS", wordinfo);
  UDM_FREE(wordinfo);
  return UDM_OK;
}


int UdmResWordInfo(UDM_ENV *Env, UDM_RESULT *Res)
{
  return UdmWWListWordInfo(&Env->Vars, &Res->WWList);
}
